#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import copy
import json
import os
from enum import Enum
from typing import List, Optional, Tuple

from parlai.core.message import Message
from parlai.core.metrics import F1Metric
from parlai.core.mutators import EpisodeMutator, MessageMutator, register_mutator
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher
from parlai.tasks.cmu_dog.build import build
from parlai.tasks.wizard_of_wikipedia.agents import (
    RareWordF1Calculator,
    TOKEN_KNOWLEDGE,
    TOKEN_END_KNOWLEDGE,
)
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager
from parlai.utils.logging import logger
from parlai.utils.typing import TShared


SILENCE = '__SILENCE__'


class SplitType(Enum):
    ORIGINAL = "original"
    ORIGINAL_DEDUPED = "deduped"
    SEEN = "seen"
    UNSEEN = "unseen"


def _datapath(opt: Opt) -> str:
    build(opt)
    return os.path.join(opt['datapath'], 'cmu_dog')


def _datafile(split: str, split_type: SplitType) -> str:
    """
    Returns the filename, e.g. train.json.
    """
    if split_type == SplitType.ORIGINAL:
        return f"{split}.json"
    if split_type == SplitType.SEEN:
        if 'test' in split:
            return "test_seen_split_seen_unseen.json"
        return f"{split}_split_seen_unseen.json"
    if split_type == SplitType.UNSEEN:
        if "test" not in split:
            logger.warning(
                "Trying to use a non-test dataset with split `unseen`. `unseen` "
                "only returns the unseen test set. Are you sure you didn't mean to "
                "use `seen` here?"
            )
        return "test_unseen_split_seen_unseen.json"
    return f"{split}_deduped.json"


def _all_split_datafiles(opt: Opt) -> List[str]:
    datafiles = []
    split_type = SplitType(opt.get("cmu_dog_split_type"))
    if split_type in {SplitType.SEEN, SplitType.UNSEEN}:
        # For seen/unseen split, the full set of dialogs is split
        # across train, valid, test seen, and test unseen
        for split in ['train', 'valid', 'test']:
            datafiles.append(_datafile(split, SplitType.SEEN))
        datafiles.append(_datafile('test', SplitType.UNSEEN))
    else:
        for split in ['train', 'valid', 'test']:
            datafiles.append(_datafile(split, split_type))
    return datafiles


def _collapse_multi_msgs(history, multi_msg_delim):
    """
    This dataset allows for a single user to send multiple messages in a row.

    Here we use a delimiter to represent this, like: "Hey!|Nice to meet you."
    """
    collapsed = []
    last_msg = history[0]
    for msg in history[1:]:
        if last_msg["uid"] == msg["uid"]:
            last_msg["text"] = multi_msg_delim.join((last_msg["text"], msg["text"]))
        else:
            collapsed.append(last_msg)
            last_msg = msg
    # don't forget to add back the last message!
    collapsed.append(last_msg)
    return collapsed


def _article_section_to_text(
    section, fact_delimiter: str, knowledge_keys: List[str] = None
) -> str:
    """
    Example input:
    {
      "cast": [
        "Ben Affleck as Batman",
        "Henry Cavill as Superman",
      ],
      "director": "Zack Snyder"
    }
    Example output:
    "cast:Ben Affleck as Batman,Henry Cavill as Superman;director:Zack Snyder"
    """
    if not section:
        return section
    if isinstance(section, str):
        return section
    texts = []
    for k, v in section.items():
        if knowledge_keys and k not in knowledge_keys:
            continue
        fact = f"{k}:"
        if isinstance(v, str):
            fact += v
        else:
            fact += ",".join(v)
        texts.append(fact)
    return fact_delimiter.join(texts)


def _build_rare_word_f1(opt: Opt) -> RareWordF1Calculator:
    datapath = _datapath(opt)

    def _collect_convo_text(convo_data):
        convo_texts = []
        for conv in convo_data.values():
            # get all messages
            convo_texts.append(' '.join([m['text'] for m in conv['history']]))
        return convo_texts

    # use conversation data from all splits for consistency
    convos = []
    convo_files = _all_split_datafiles(opt)

    for fname in convo_files:
        with PathManager.open(os.path.join(datapath, f"conversations/{fname}")) as f:
            data = json.load(f)
        convos += _collect_convo_text(data)

    return RareWordF1Calculator(' '.join(convos), top_p=0.5)


class CMUDocumentGroundedConversationsTeacher(DialogTeacher):
    """
    CMU Document Grounded Conversations Dataset (aka CMU_DoG)

    Paper: https://arxiv.org/pdf/1809.07358.pdf
    Source: https://github.com/festvox/datasets-CMU_DoG
    """

    def __init__(self, opt: Opt, shared: TShared = None):
        opt = copy.deepcopy(opt)
        self.delimiter = opt.get('delimiter', '\n')
        split_type = SplitType(opt.get("cmu_dog_split_type"))
        if split_type == SplitType.ORIGINAL:
            logger.warning(
                "`original` split type contains duplicate conversations across train, "
                "valid, and test. See https://github.com/festvox/datasets-CMU_DoG/issues/2 "
                "for more detail."
            )
        opt['datafile'] = _datafile(
            split=DatatypeHelper.fold(opt['datatype']), split_type=split_type
        )
        super().__init__(opt, shared)
        if shared:
            self.rare_word_f1 = shared['rare_word_f1']
        else:
            self.rare_word_f1 = _build_rare_word_f1(opt)

    @classmethod
    def add_cmdline_args(
        cls, parser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        cmu_dog = parser.add_argument_group("CMU Document Grounded Conversations")
        cmu_dog.add_argument(
            "--cmu-dog-rating",
            type=int,
            default=[1, 2, 3],
            choices=[1, 2, 3],
            nargs="+",
            help='The higher the number, the better quality the conversation. '
            'For each rating, the number of conversations is as follows: 1-1443, 2-2142, 3-527',
        )
        cmu_dog.add_argument(
            "--cmu-dog-only-with-knowledge",
            type=bool,
            default=True,
            help="Optionally train only the sides of the conversation that have access to knowledge.",
        )
        cmu_dog.add_argument(
            "--cmu-dog-multi-msg-delimiter",
            type=str,
            default=" ",
            help="When one agent is to send multiple messages in a row, they will be concatenated with this delimiter.",
        )
        cmu_dog.add_argument(
            "--cmu-dog-fact-delimiter",
            type=str,
            default=";",
            help="When a section of the knowledge contains multiple facts, they will be concatenated with this delimiter.",
        )
        cmu_dog.add_argument(
            "--cmu-dog-include-knowledge-keys",
            type=str,
            default='cast,critical_response,director,genre,introduction,movieName,rating,year',
            help="Comma-separated list of keys into the knowledge to include as the general conversational context",
        )
        cmu_dog.add_argument(
            "--cmu-dog-provide-movie-context",
            type=bool,
            default=True,
            help="Provide movie facts as the general conversational context",
        )
        cmu_dog.add_argument(
            "--cmu-dog-split-type",
            type=SplitType,
            default=SplitType.ORIGINAL_DEDUPED,
            choices=list(SplitType),
            help=(
                "`orginal`: train/valid/test split from the original release, "
                "`original_deduped`: duplicate conversations removed from train set, "
                "`seen`: refers to movies and is relative to training - `test seen` is conversations about movies that appear during training, "
                "`unseen`: contains conversations about movies that weren't seen in `train`/`valid`/`test seen`. "
                "When using seen/unseen, use `seen` for `train`/`valid`/`seen test` and `unseen` only for `unseen test`."
            ),
        )
        return parser

    def share(self):
        shared = super().share()
        shared['rare_word_f1'] = self.rare_word_f1
        return shared

    def setup_data(self, datafile: str):
        datapath = _datapath(self.opt)
        with PathManager.open(os.path.join(datapath, f"conversations/{datafile}")) as f:
            data = json.load(f)
        with PathManager.open(os.path.join(datapath, "wiki_data.json")) as f:
            wiki_data = json.load(f)

        # Filter by rating
        data = {
            k: c for k, c in data.items() if c["rating"] in self.opt["cmu_dog_rating"]
        }

        def _can_see_info(turn, convo):
            # Sometimes only one participant has access to the article
            return turn["uid"] in convo["whoSawDoc"]

        num_eps = len(data)
        data = list(data.items())
        # loop through conversations
        for i in range(len(data) * 2):
            conv_idx = i % num_eps
            start_idx = i // num_eps

            _conv_id, conv_data = data[conv_idx]

            dialog = _collapse_multi_msgs(
                conv_data["history"], self.opt['cmu_dog_multi_msg_delimiter']
            )
            movie_article = wiki_data[str(conv_data["wikiDocumentIdx"])]

            if self.opt["cmu_dog_only_with_knowledge"] and not _can_see_info(
                dialog[start_idx], conv_data
            ):
                continue

            # loop through turns
            for idx in range(start_idx, len(dialog), 2):
                label_turn = dialog[idx]
                label = label_turn["text"].strip()

                # The section displayed changes across the conversation
                doc_idx = str(label_turn["docIdx"])
                gold_knowledge = _article_section_to_text(
                    movie_article[doc_idx], self.opt['cmu_dog_fact_delimiter']
                )
                section = (
                    movie_article[doc_idx]
                    if _can_see_info(label_turn, conv_data)
                    else None
                )
                section_text = _article_section_to_text(
                    section,
                    self.opt['cmu_dog_fact_delimiter'],
                    self.opt.get('cmu_dog_include_knowledge_keys').split(','),
                )

                # By default, start conversation with silence
                if idx == start_idx:
                    context = (
                        section_text
                        if self.opt['cmu_dog_provide_movie_context']
                        else SILENCE
                    )
                else:
                    context = dialog[idx - 1]["text"].strip()

                yield Message(
                    {
                        'text': context,
                        'labels': [label],
                        'available_knowledge_raw': section,
                        'available_knowledge_text': section_text,
                        'title': movie_article['0']['movieName'],
                        'checked_sentence': gold_knowledge,
                    }
                ), idx == start_idx

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ):
        if 'text' in model_response and 'checked_sentence' in teacher_action:
            self.metrics.add(
                'knowledge_f1',
                F1Metric.compute(
                    model_response['text'], [teacher_action['checked_sentence']]
                ),
            )
        if 'text' in model_response and labels:
            self.metrics.add(
                'rare_word_f1',
                self.rare_word_f1.compute(model_response['text'], labels),
            )


@register_mutator("prepend_knowledge_to_message")
class PrependKnowledgeToMessageMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        if not message.get('available_knowledge_text'):
            return message
        context = message.pop('text')
        knowledge = f'{TOKEN_KNOWLEDGE} {message["available_knowledge_text"]} {TOKEN_END_KNOWLEDGE}'
        delimiter = self.opt.get('delimiter', '\n')
        message['text'] = (
            knowledge if context == SILENCE else f'{knowledge}{delimiter}{context}'
        )
        return message


@register_mutator("knowledge_only_when_updated")
class KnowledgeWhenUpdatedMutator(EpisodeMutator):
    def episode_mutation(self, episode: List[Message]) -> List[Message]:
        last_knowledge = None
        for msg in episode:
            knowledge = msg.pop('available_knowledge_text')
            if last_knowledge != knowledge:
                msg['available_knowledge_text'] = knowledge
            last_knowledge = knowledge
        return episode


class DefaultTeacher(CMUDocumentGroundedConversationsTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
APIs for retrieving a list of "Contents" using an "Search Query".

The term "Search Query" here refers to any abstract form of input string. The definition
of "Contents" is also loose and depends on the API.
"""

from abc import ABC, abstractmethod
import requests
from typing import Any, Dict, List

from parlai.core.opt import Opt
from parlai.utils import logging


CONTENT = 'content'
DEFAULT_NUM_TO_RETRIEVE = 5


class RetrieverAPI(ABC):
    """
    Provides the common interfaces for retrievers.

    Every retriever in this modules must implement the `retrieve` method.
    """

    def __init__(self, opt: Opt):
        self.skip_query_token = opt['skip_retrieval_token']

    @abstractmethod
    def retrieve(
        self, queries: List[str], num_ret: int = DEFAULT_NUM_TO_RETRIEVE
    ) -> List[Dict[str, Any]]:
        """
        Implements the underlying retrieval mechanism.
        """

    def create_content_dict(self, content: list, **kwargs) -> Dict:
        resp_content = {CONTENT: content}
        resp_content.update(**kwargs)
        return resp_content


class SearchEngineRetrieverMock(RetrieverAPI):
    """
    For unit tests and debugging (does not need a running server).
    """

    def retrieve(
        self, queries: List[str], num_ret: int = DEFAULT_NUM_TO_RETRIEVE
    ) -> List[Dict[str, Any]]:
        all_docs = []
        for query in queries:
            if query == self.skip_query_token:
                docs = None
            else:
                docs = []
                for idx in range(num_ret):
                    doc = self.create_content_dict(
                        f'content {idx} for query "{query}"',
                        url=f'url_{idx}',
                        title=f'title_{idx}',
                    )
                    docs.append(doc)
            all_docs.append(docs)
        return all_docs


class SearchEngineRetriever(RetrieverAPI):
    """
    Queries a server (eg, search engine) for a set of documents.

    This module relies on a running HTTP server. For each retrieval it sends the query
    to this server and receives a JSON; it parses the JSON to create the the response.
    """

    def __init__(self, opt: Opt):
        super().__init__(opt=opt)
        self.server_address = self._validate_server(opt.get('search_server'))

    def _query_search_server(self, query_term, n):
        server = self.server_address
        req = {'q': query_term, 'n': n}
        logging.debug(f'sending search request to {server}')
        server_response = requests.post(server, data=req)
        resp_status = server_response.status_code
        if resp_status == 200:
            return server_response.json().get('response', None)
        logging.error(
            f'Failed to retrieve data from server! Search server returned status {resp_status}'
        )

    def _validate_server(self, address):
        if not address:
            raise ValueError('Must provide a valid server for search')
        if address.startswith('http://') or address.startswith('https://'):
            return address
        PROTOCOL = 'http://'
        logging.warning(f'No portocol provided, using "{PROTOCOL}"')
        return f'{PROTOCOL}{address}'

    def _retrieve_single(self, search_query: str, num_ret: int):
        if search_query == self.skip_query_token:
            return None

        retrieved_docs = []
        search_server_resp = self._query_search_server(search_query, num_ret)
        if not search_server_resp:
            logging.warning(
                f'Server search did not produce any results for "{search_query}" query.'
                ' returning an empty set of results for this query.'
            )
            return retrieved_docs

        for rd in search_server_resp:
            url = rd.get('url', '')
            title = rd.get('title', '')
            sentences = [s.strip() for s in rd[CONTENT].split('\n') if s and s.strip()]
            retrieved_docs.append(
                self.create_content_dict(url=url, title=title, content=sentences)
            )
        return retrieved_docs

    def retrieve(
        self, queries: List[str], num_ret: int = DEFAULT_NUM_TO_RETRIEVE
    ) -> List[Dict[str, Any]]:
        # TODO: update the server (and then this) for batch responses.
        return [self._retrieve_single(q, num_ret) for q in queries]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Utilities related to handling data.
"""
import random
from typing import List


class DatatypeHelper:
    """
    Helper class to determine properties from datatype strings.
    """

    @classmethod
    def fold(cls, datatype: str) -> str:
        """
        Extract the fold part of the datatype.

        :param datatype:
            parlai datatype

        :return: the fold

        >>> DatatypeHelper.fold("train:ordered")
        ... "train"
        """
        return datatype.split(':')[0]

    @classmethod
    def strip_stream(cls, datatype: str) -> str:
        """
        Remove :stream from the datatype.

        Used by ChunkTeacher where behavior does not change based on streaming.

        :param datatype:
            parlai datatype

        :return:
            a non-streaming version of the datatype.

        >>> DatatypeHelper.fold("train:stream")
        "train"
        >>> DatatypeHelper.fold("train")
        "train"
        """
        return datatype.replace(":stream", "")

    @classmethod
    def should_cycle(cls, datatype: str) -> bool:
        """
        Return whether we should cycle data based on the datatype.

        :param datatype:
            parlai datatype

        :return should_cycle:
            given datatype, return whether we should cycle
        """
        assert datatype is not None, 'datatype must not be none'
        return (
            'train' in datatype
            and 'evalmode' not in datatype
            and 'ordered' not in datatype
        )

    @classmethod
    def should_shuffle(cls, datatype: str) -> bool:
        """
        Return whether we should shuffle data based on the datatype.

        :param datatype:
            parlai datatype

        :return should_shuffle:
            given datatype, return whether we should shuffle
        """
        assert datatype is not None, 'datatype must not be none'
        return (
            'train' in datatype
            and 'evalmode' not in datatype
            and 'ordered' not in datatype
            and 'stream' not in datatype
        )

    @classmethod
    def is_training(cls, datatype: str) -> bool:
        """
        Return whether we should return eval_labels or labels.

        :param datatype:
            parlai datatype

        :return is_training:
            bool indicating whether should return eval_labels or labels
        """
        assert datatype is not None, 'datatype must not be none'
        return 'train' in datatype and 'evalmode' not in datatype

    @classmethod
    def is_streaming(cls, datatype: str) -> bool:
        """
        Return whether this is streaming.

        :param datatype:
            parlai datatype

        :returns:
            bool indicating whether we are streaming
        """
        return 'stream' in datatype

    @classmethod
    def split_data_by_fold(
        cls,
        fold: str,
        data: List,
        train_frac: float,
        valid_frac: float,
        test_frac: float,
        seed: int = 42,
    ):
        """
        Splits a list of data into train/valid/test folds. The members of these folds
        are randomized (in a consistent manner) by a seed. This is a convenience
        function for datasets that do not have a canonical split.

        :param fold:
           parlai fold/datatype
        :param data:
            List of data examples to be split
        :param train_frac:
            Fraction of data to be used for the "train" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param valid_frac:
            Fraction of data to be used for the "valid" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param test_frac:
            Fraction of data to be used for the "test" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param seed:
            Seed for shuffling
        """
        assert train_frac + valid_frac + test_frac == 1
        if "train" in fold:
            start = 0.0
            end = train_frac
        elif "valid" in fold:
            start = train_frac
            end = train_frac + valid_frac
        else:
            start = train_frac + valid_frac
            end = 1.0

        random.Random(seed).shuffle(data)
        return data[int(start * len(data)) : int(end * len(data))]

    @classmethod
    def split_subset_data_by_fold(
        cls,
        fold: str,
        subsets: List[List],
        train_frac: float,
        valid_frac: float,
        test_frac: float,
        seed: int = 42,
    ):
        """
        Splits a list of subsets of data, where we want equal samples from each subset,
        into train/valid/test folds, ensuring that samples from a given subset are not
        changed to another fold as more subsets are added.

        For example, say a dataset has domains A, B. Let's say we have an experiment where we train and validate a model on domain A, then on domains A + B. If we naively concatinate the subset of data from A + B together, randomize it, and split the result into train, valid, and test folds, there is no guarantee that valid or test examples from A-only will not end up into the train fold of the A + B split from this naive concatination process.

        The members of these folds are randomized (but in a fixed manner) by a seed.

        :param fold:
           parlai fold/datatype
        :param subsets:
            List of subsets of data examples to be split
        :param train_frac:
            Fraction of data to be used for the "train" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param valid_frac:
            Fraction of data to be used for the "valid" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param test_frac:
            Fraction of data to be used for the "test" fold. train_frac, valid_frac, and test_frac should sum to 1.
        :param seed:
            Seed for shuffling
        """
        result = []
        for subset in subsets:
            result.extend(
                cls.split_data_by_fold(
                    fold, subset, train_frac, valid_frac, test_frac, seed
                )
            )
        random.Random(seed).shuffle(result)
        return result



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SubModule code for BlenderBot2.

Contains implementations of the Query Generator and Memory Decoder.
"""
from enum import Enum, auto
import os
import string
import time
import torch
import torch.nn
from typing import List, Tuple, Dict, Optional, Any

from parlai.agents.rag.retrievers import clean_vec
from parlai.core.agents import create_agent_from_model_file, create_agent_from_shared
from parlai.core.build_data import modelzoo_path
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.torch_agent import TorchAgent
from parlai.tasks.msc.agents import NOPERSONA
import parlai.utils.logging as logging


# The below are the classification outputs mapping to either
# retrieving from memory, or not retrieving at all.
MEMORY_STRINGS = ['convai2', 'personal_knowledge']
NONE_STRINGS = [
    'blended_skill_talk',
    'empathetic_dialogues',
    'dummy',
    'no_passages_used',
]


def strip_punc(s):
    return s.translate(str.maketrans('', '', string.punctuation.replace('_', '')))


def clean_vec_with_dict(dict: DictionaryAgent, vec: torch.LongTensor) -> List[int]:
    """
    Clean the specified vector with the specified dictionary.

    See `parlai.agents.rag.retrievers.clean_vec`for a description
    """
    return clean_vec(
        vec,
        dict[dict.end_token],  # type: ignore
        special_toks=[
            dict[dict.null_token],  # type: ignore
            dict[dict.start_token],  # type: ignore
            dict[dict.end_token],  # type: ignore
            dict[dict.unk_token],  # type: ignore
        ],
    )


class RetrievalType(Enum):
    """
    Retrieval Type indicates the "type" of retrieval.

    That is, we either don't retrieve; retrieve from memory; or retrieve via search.
    """

    NONE = auto()
    SEARCH = auto()
    MEMORY = auto()


class KnowledgeAccessMethod(Enum):
    """
    How BlenderBot2 should retrieve for each input.

    classify => classify the input text, determine which retrieval to use

    memory_only => only retrieve via memories (i.e., from dialogue context)
    search_only => only retrieve via internet/FAISS search
    all => for each input, retrieve both from memories and internet/FAISS search
    none => do not retrieve anything.
    """

    CLASSIFY = 'classify'
    MEMORY_ONLY = 'memory_only'
    SEARCH_ONLY = 'search_only'
    ALL = 'all'
    NONE = 'none'


class BB2SubmoduleMixin:
    """
    Mixin for agents used within BB2.

    agents: list of agents
    agent_dict: dictionary for the agent
    input_type: for logging purposes.
    """

    agents: List[TorchAgent]
    agent_dict: Optional[DictionaryAgent]
    input_type: str
    generations: List[str]

    def tokenize_input(self, input: str) -> List[int]:
        """
        Tokenize input for the sub agent.

        Assumes that the sub agent has been instantiated.

        :param input:
            input to the sub agent

        :return tokens:
            return tokenized input
        """
        assert self.agents and self.agent_dict is not None
        return self.agent_dict.txt2vec(input)

    def clean_input(self, vec: torch.LongTensor) -> List[int]:
        """
        Clean a tensor before converting to a string.
        """
        assert self.agent_dict is not None
        return clean_vec_with_dict(self.agent_dict, vec)

    def _batch_generate(self, texts: List[str]) -> List[str]:
        """
        Batch generate items from an input list of texts.

        :param texts:
            list of texts

        :return generations:
            return agent generations for each input.
        """
        start = time.time()
        active_agents = self.agents[: len(texts)]
        for agent_i, t_i in zip(active_agents, texts):
            agent_i.observe(Message({'text': t_i, 'episode_done': True}))
        agent_replies = self.agents[0].batch_act([a.observation for a in active_agents])
        logging.debug(f'Generated: {time.time() - start:.2f}')
        for agent_i, reply_i in zip(active_agents, agent_replies):
            agent_i.self_observe(reply_i)
        self.generations = [r.get('text', 'dummy') for r in agent_replies]
        return self.generations


class QueryGenerator(BB2SubmoduleMixin):
    """
    The QueryGenerator is a wrapper around a generator model.

    This model can be trained for both dataset classification and search query
    generation.
    """

    def __init__(self, opt: Opt):
        self.opt = opt
        self.agents = []
        self.agent_dict = None
        self.generations = []
        self.input_type = 'Search'
        self.knowledge_access_method = KnowledgeAccessMethod(
            opt['knowledge_access_method']
        )
        model_file = modelzoo_path(opt['datapath'], opt['query_generator_model_file'])
        if model_file and os.path.exists(model_file):
            logging.info(f'Building Query Generator from file: {model_file}')
            logging.disable()
            overrides: Dict[str, Any] = {'skip_generation': False}
            overrides['inference'] = opt['query_generator_inference']
            overrides['beam_size'] = opt.get('query_generator_beam_size', 3)
            overrides['beam_min_length'] = opt.get('query_generator_beam_min_length', 2)
            if self.opt['query_generator_truncate'] > 0:
                overrides['text_truncate'] = self.opt['query_generator_truncate']
                overrides['truncate'] = self.opt['query_generator_truncate']
            base_agent = create_agent_from_model_file(
                model_file, opt_overrides=overrides
            )
            assert isinstance(base_agent, TorchAgent)
            self.agents = [base_agent]
            bsz = opt.get('batchsize', 1)
            rag_turn_n_turns = opt.get('rag_turn_n_turns', 1)
            if bsz > 1 or rag_turn_n_turns > 1:
                self.agents += [
                    create_agent_from_shared(self.agents[0].share())
                    for _ in range((bsz * rag_turn_n_turns) - 1)
                ]
            self.agent_dict = self.agents[0].build_dictionary()
            logging.enable()

    def classify_retrieval(
        self,
        input: torch.LongTensor,
        num_memories: torch.LongTensor,
        generated_memories: Optional[List[List[str]]],
    ) -> Tuple[torch.LongTensor, List[str]]:
        """
        Classify input and get retrieval type.

        Here, we classify which "type" of retrieval to do for each input batch item.

        In the case of "search", we additionally return search queries.

        :param input:
            input to classify
        :param num_memories:
            how many memories each example has.
            we override classification if there are no mems for the example.
        :param generated_memories:
            the generated memories from a memory decoder.

        :return (retrieval_type, searches):
            retrieval_type: a bsz-length tensor indicating which "type" of retrieval
                            we're doing (see RetrievalType above)
            searches: For batch items classified as search, we return the search queries
                      as well.
        """
        self.retrieval_type = torch.LongTensor(input.size(0))
        self.retrieval_type.fill_(0)
        assert self.agent_dict is not None
        texts = [self.agent_dict.vec2txt(self.clean_input(i)) for i in input]
        if self.knowledge_access_method is KnowledgeAccessMethod.MEMORY_ONLY:
            search_queries = [MEMORY_STRINGS[-1]] * len(texts)
        else:
            search_queries = self._batch_generate(texts)
        logging.debug(f'search queries: {search_queries}')
        logging.verbose(f'Search: {search_queries[0]}')
        searches = []
        if not generated_memories:
            generated_memories = [[] for _ in range(input.size(0))]
        for i, s in enumerate(search_queries):
            if (
                (strip_punc(s) in MEMORY_STRINGS)
                or any(ms in s for ms in MEMORY_STRINGS)
            ) and (
                (num_memories is not None and num_memories[i] > 0)
                or generated_memories[i]
            ):
                self.retrieval_type[i] = RetrievalType.MEMORY.value
            elif strip_punc(s) in NONE_STRINGS + MEMORY_STRINGS:
                self.retrieval_type[i] = RetrievalType.NONE.value
            else:
                self.retrieval_type[i] = RetrievalType.SEARCH.value
                searches.append(s)

        return self.retrieval_type, searches


class MemoryDecoder(BB2SubmoduleMixin):
    """
    Memory decoder.

    Given a line of context input, generate a memory to write.
    """

    def __init__(self, opt: Opt):
        self.opt = opt
        self.agents = []
        self.agent_dict = None
        self.generations = []
        self.input_type = 'Memory'
        self.delimiter = opt.get('memory_decoder_delimiter', '\n')
        self.one_line_memories = opt.get('memory_decoder_one_line_memories', False)
        model_file = modelzoo_path(opt['datapath'], opt['memory_decoder_model_file'])
        if model_file and os.path.exists(model_file):
            logging.info(f'Building Memory Decoder from file: {model_file}')
            logging.disable()
            overrides = {
                'skip_generation': False,
                'inference': 'beam',
                'beam_size': opt.get('memory_decoder_beam_size', 3),
                'beam_min_length': opt.get('memory_decoder_beam_min_length', 10),
                'beam_block_ngram': 3,
            }
            if self.opt.get('memory_decoder_truncate', -1) > 0:
                overrides['text_truncate'] = self.opt['memory_decoder_truncate']
                overrides['truncate'] = self.opt['memory_decoder_truncate']
            base_agent = create_agent_from_model_file(
                model_file, opt_overrides=overrides
            )
            assert isinstance(base_agent, TorchAgent)
            self.agents = [base_agent]
            assert isinstance(self.agents[0], TorchAgent)
            copies = max(100, (opt['batchsize'] * opt.get('rag_turn_n_turns', 1)))
            self.agents += [
                create_agent_from_shared(self.agents[0].share()) for _ in range(copies)
            ]
            self.agent_dict = self.agents[0].build_dictionary()
            logging.enable()

    def generate_memories(
        self, input: torch.LongTensor, num_inputs: torch.LongTensor
    ) -> List[List[str]]:
        """
        Generate memories from input.

        Each input is split into the lines of conversational context.
        These are considered independently.

        We then assign a prefix ("your/partner's persona:") dependent on
        whether the bot or it's partner said the line.

        :param input:
            input to the memory decoder
        :param num_inputs:
            number of lines per batch item
        """
        assert self.agent_dict is not None
        memories = []
        offset = 0
        for idx, i in enumerate(input):
            if num_inputs[idx] == 0:
                continue
            context_lines_vec = i[offset : offset + num_inputs[idx]]
            offset += num_inputs[idx]
            context_lines = [
                self.agent_dict.vec2txt(self.clean_input(j)) for j in context_lines_vec
            ]
            raw_memories_i = list(reversed(self._batch_generate(context_lines)))
            logging.debug(f'raw memories: {raw_memories_i}')
            memories_i = self._extract_from_raw_memories(raw_memories_i)
            logging.debug(f'memories to write: {memories_i}')
            mem_string = '\n'.join(memories_i)
            logging.verbose(f'Writing memories: {mem_string}')
            memories.append(memories_i)
        return memories

    def _extract_from_raw_memories(self, raw_memories: List[str]) -> List[str]:
        """
        Extract memory lines from batch generated memories.

        Prefixes accordingly, and combines on one line if necessary.

        :param raw_memories:
            raw memory generations. sometimes we need skip the memories because
            nothing was generated

        :return memories:
            return prefixed and filtered memories
        """
        partner_prefix = 'partner\'s persona:'
        self_prefix = 'your persona:'
        num_ctxt = len(raw_memories)
        memories = []
        partner_memories = []
        self_memories = []
        for idx in range(num_ctxt):
            if raw_memories[idx] == NOPERSONA:
                continue
            if idx % 2 == 0:
                partner_memories.append(raw_memories[idx])
                prefix = partner_prefix
            else:
                self_memories.append(raw_memories[idx])
                prefix = self_prefix
            if not self.one_line_memories:
                memories.append(f'{prefix} {raw_memories[idx]}')
        if self.one_line_memories:
            if partner_memories:
                memories.append(f"{partner_prefix} {' '.join(partner_memories)}")
            if self_memories:
                memories.append(f"{self_prefix} {' '.join(self_memories)}")

        return memories



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Torch Classifier Agents classify text into a fixed set of labels.
"""


from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.utils.torch import PipelineHelper, total_parameters, trainable_parameters
from parlai.core.torch_agent import TorchAgent, Output
from parlai.utils.misc import round_sigfigs, warn_once
from parlai.core.metrics import Metric, AverageMetric
from typing import List, Optional, Tuple, Dict, Union
from typing import Counter
from parlai.utils.typing import TScalar
from parlai.utils.io import PathManager
from sklearn.metrics import auc

import parlai.utils.logging as logging
import torch.nn.functional as F
import torch
import math


class ConfusionMatrixMetric(Metric):
    """
    Class that keeps count of the confusion matrix for classification.

    Also provides helper methods computes precision, recall, f1, weighted_f1 for
    classification.
    """

    __slots__ = (
        '_true_positives',
        '_true_negatives',
        '_false_positives',
        '_false_negatives',
    )

    @property
    def macro_average(self) -> bool:
        """
        Indicates whether this metric should be macro-averaged when globally reported.
        """
        return True

    def __init__(
        self,
        true_positives: TScalar = 0,
        true_negatives: TScalar = 0,
        false_positives: TScalar = 0,
        false_negatives: TScalar = 0,
    ) -> None:
        self._true_positives = self.as_number(true_positives)
        self._true_negatives = self.as_number(true_negatives)
        self._false_positives = self.as_number(false_positives)
        self._false_negatives = self.as_number(false_negatives)

    def __add__(
        self, other: Optional['ConfusionMatrixMetric']
    ) -> 'ConfusionMatrixMetric':
        # NOTE: hinting can be cleaned up with "from __future__ import annotations" when
        # we drop Python 3.6
        if other is None:
            return self
        assert isinstance(other, ConfusionMatrixMetric)
        full_true_positives: TScalar = self._true_positives + other._true_positives
        full_true_negatives: TScalar = self._true_negatives + other._true_negatives
        full_false_positives: TScalar = self._false_positives + other._false_positives
        full_false_negatives: TScalar = self._false_negatives + other._false_negatives

        # always keep the same return type
        return type(self)(
            true_positives=full_true_positives,
            true_negatives=full_true_negatives,
            false_positives=full_false_positives,
            false_negatives=full_false_negatives,
        )

    @staticmethod
    def compute_many(
        true_positives: TScalar = 0,
        true_negatives: TScalar = 0,
        false_positives: TScalar = 0,
        false_negatives: TScalar = 0,
    ) -> Tuple['PrecisionMetric', 'RecallMetric', 'ClassificationF1Metric']:
        return (
            PrecisionMetric(
                true_positives, true_negatives, false_positives, false_negatives
            ),
            RecallMetric(
                true_positives, true_negatives, false_positives, false_negatives
            ),
            ClassificationF1Metric(
                true_positives, true_negatives, false_positives, false_negatives
            ),
        )

    @staticmethod
    def compute_metrics(
        predictions: List[str], gold_labels: List[str], positive_class: str
    ) -> Tuple[
        List['PrecisionMetric'], List['RecallMetric'], List['ClassificationF1Metric']
    ]:
        precisions = []
        recalls = []
        f1s = []
        for predicted, gold_label in zip(predictions, gold_labels):
            true_positives = int(
                predicted == positive_class and gold_label == positive_class
            )
            true_negatives = int(
                predicted != positive_class and gold_label != positive_class
            )
            false_positives = int(
                predicted == positive_class and gold_label != positive_class
            )
            false_negatives = int(
                predicted != positive_class and gold_label == positive_class
            )
            precision, recall, f1 = ConfusionMatrixMetric.compute_many(
                true_positives, true_negatives, false_positives, false_negatives
            )
            precisions.append(precision)
            recalls.append(recall)
            f1s.append(f1)
        return precisions, recalls, f1s


class PrecisionMetric(ConfusionMatrixMetric):
    """
    Class that takes in a ConfusionMatrixMetric and computes precision for classifier.
    """

    def value(self) -> float:
        if self._true_positives == 0:
            return 0.0
        else:
            return self._true_positives / (self._true_positives + self._false_positives)


class RecallMetric(ConfusionMatrixMetric):
    """
    Class that takes in a ConfusionMatrixMetric and computes recall for classifier.
    """

    def value(self) -> float:
        if self._true_positives == 0:
            return 0.0
        else:
            return self._true_positives / (self._true_positives + self._false_negatives)


class ClassificationF1Metric(ConfusionMatrixMetric):
    """
    Class that takes in a ConfusionMatrixMetric and computes f1 for classifier.
    """

    def value(self) -> float:
        if self._true_positives == 0:
            return 0.0
        else:
            numer = 2 * self._true_positives
            denom = numer + self._false_negatives + self._false_positives
            return numer / denom


class AUCMetrics(Metric):
    """
    Computes Area Under ROC Curve (AUC) metrics.

    Does so by keeping track of positives' and negatives' probability score counts in
    Counters or dictionaries. Note the introduction of `max_bucket_dec_places`; this
    integer number determines the number of digits to save for the probability scores. A
    higher `max_bucket_dec_places` will a more accurate estimate of the exact AUC
    metric, but may also use more space.

    NOTE: currently only used for classifiers in the `eval_model` script; to use,
    add the argument `-auc <max_bucket_dec_places>` when calling `eval_model` script
    """

    @property
    def macro_average(self) -> bool:
        """
        Indicates whether this metric should be macro-averaged when globally reported.
        """
        return False

    @classmethod
    def raw_data_to_auc(
        cls,
        true_labels: List[Union[int, str]],
        pos_probs: List[float],
        class_name,
        max_bucket_dec_places: int = 3,
    ):
        auc_object = cls(class_name, max_bucket_dec_places=max_bucket_dec_places)
        auc_object.update_raw(
            true_labels=true_labels, pos_probs=pos_probs, class_name=class_name
        )
        return auc_object

    def __init__(
        self,
        class_name: Union[int, str],
        max_bucket_dec_places: int = 3,
        pos_dict: Optional[Counter[float]] = None,
        neg_dict: Optional[Counter[float]] = None,
    ):
        # `_pos_dict` keeps track of the probabilities of the positive class
        self._pos_dict = pos_dict if pos_dict else Counter()
        # `_neg_dict` keeps track of the probabilities of the negative class
        self._neg_dict = neg_dict if neg_dict else Counter()
        self._class_name = class_name
        self._max_bucket_dec_places = max_bucket_dec_places

    def update_raw(
        self, true_labels: List[Union[int, str]], pos_probs: List[float], class_name
    ):
        """
        given the true/golden labels and the probabilities of the positive class, we
        will update our bucket dictionaries of positive and negatives (based on the
        class_name); `max_bucket_dec_places` is also used here to round the
        probabilities and possibly.
        """
        assert self._class_name == class_name
        assert len(true_labels) == len(pos_probs)

        TO_INT_FACTOR = 10 ** self._max_bucket_dec_places
        # add the upper and lower bound of the values
        for label, prob in zip(true_labels, pos_probs):
            # calculate the upper and lower bound of the values
            prob_down = math.floor(prob * TO_INT_FACTOR) / TO_INT_FACTOR
            if label == self._class_name:
                interested_dict = self._pos_dict
            else:
                interested_dict = self._neg_dict
            if interested_dict.get(prob_down):
                interested_dict[prob_down] += 1
            else:
                interested_dict[prob_down] = 1

    def __add__(self, other: Optional['AUCMetrics']) -> 'AUCMetrics':
        if other is None:
            return self
        assert isinstance(other, AUCMetrics)
        assert other._class_name == self._class_name
        all_pos_dict = self._pos_dict + other._pos_dict
        all_neg_dict = self._neg_dict + other._neg_dict

        return AUCMetrics(
            self._class_name, pos_dict=all_pos_dict, neg_dict=all_neg_dict
        )

    def _calc_fp_tp(self) -> List[Tuple[int]]:
        """
        Calculates the False Positives and True positives; returned as a list of pairs:

        `[(fp, tp)]`
        """
        all_thresholds = sorted(
            set(list(self._pos_dict.keys()) + list(self._neg_dict.keys()))
        )
        # sorted in ascending order,
        # so adding a upper bound so that its tp, fp is (0, 0)
        all_thresholds.append(all_thresholds[-1] + 1)
        L = len(all_thresholds)
        # false positives, true positives
        fp_tp = [(0, 0)]

        # the biggest one is always (0,0), so skip that one
        for i in range(L - 2, -1, -1):
            fp, tp = fp_tp[-1]
            thres = all_thresholds[i]
            # false positives
            fp += self._neg_dict.get(thres, 0)
            # true positives
            tp += self._pos_dict.get(thres, 0)
            fp_tp.append((fp, tp))
        return fp_tp

    def _calc_fpr_tpr(self) -> Tuple[Union[List[int], int]]:
        """
        Calculates the false positive rates and true positive rates Also returns the
        total number of positives and negatives; returned as a list of pairs and two
        integers:

        `([(fpr, tpr)], positives, negatives)`; note that if the total
        negatives/positives is 0, then will return 0 for either fpr/tpr instead of
        raising an error
        """
        _tot_pos = sum(self._pos_dict.values())
        _tot_neg = sum(self._neg_dict.values())
        fp_tp = self._calc_fp_tp()
        fps, tps = list(zip(*fp_tp))
        if _tot_neg == 0:
            fpr = [0] * len(fps)
        else:
            fpr = [fp / _tot_neg for fp in fps]

        if _tot_pos == 0:
            tpr = [0] * len(tps)
        else:
            tpr = [tp / _tot_pos for tp in tps]

        return (list(zip(fpr, tpr)), _tot_pos, _tot_neg)

    def value(self) -> float:
        fpr_tpr, _tot_pos, _tot_neg = self._calc_fpr_tpr()

        if _tot_pos == 0 and _tot_neg == 0:
            return 0

        # auc needs x-axis to be sorted
        fpr_tpr.sort()
        fpr, tpr = list(zip(*fpr_tpr))
        return auc(fpr, tpr)


class WeightedF1Metric(Metric):
    """
    Class that represents the weighted f1 from ClassificationF1Metric.
    """

    __slots__ = '_values'

    @property
    def macro_average(self) -> bool:
        """
        Indicates whether this metric should be macro-averaged when globally reported.
        """
        return True

    def __init__(self, metrics: Dict[str, ClassificationF1Metric]) -> None:
        self._values: Dict[str, ClassificationF1Metric] = metrics

    def __add__(self, other: Optional['WeightedF1Metric']) -> 'WeightedF1Metric':
        if other is None:
            return self
        assert isinstance(other, WeightedF1Metric)
        output: Dict[str, ClassificationF1Metric] = dict(**self._values)
        for k, v in other._values.items():
            output[k] = output.get(k, None) + v  # type: ignore
        return WeightedF1Metric(output)

    def value(self) -> float:
        weighted_f1 = 0.0
        values = list(self._values.values())
        if len(values) == 0:
            return weighted_f1
        total_examples = sum(
            [each._true_positives + each._false_negatives for each in values]
        )
        for each in values:
            actual_positive = each._true_positives + each._false_negatives
            weighted_f1 += each.value() * (actual_positive / total_examples)
        return weighted_f1

    @staticmethod
    def compute_many(
        metrics: Dict[str, List[ClassificationF1Metric]]
    ) -> List['WeightedF1Metric']:
        weighted_f1s = [dict(zip(metrics, t)) for t in zip(*metrics.values())]
        return [WeightedF1Metric(metrics) for metrics in weighted_f1s]


class TorchClassifierAgent(TorchAgent):
    """
    Abstract Classifier agent. Only meant to be extended.

    TorchClassifierAgent aims to handle much of the bookkeeping any classification
    model.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add CLI args.
        """
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        parser = parser.add_argument_group('Torch Classifier Arguments')
        # class arguments
        parser.add_argument(
            '--classes',
            type=str,
            nargs='*',
            default=None,
            help='the name of the classes.',
        )
        parser.add_argument(
            '--class-weights',
            type=float,
            nargs='*',
            default=None,
            help='weight of each of the classes for the softmax',
        )
        parser.add_argument(
            '--ref-class',
            type=str,
            default=None,
            hidden=True,
            help='the class that will be used to compute '
            'precision and recall. By default the first '
            'class.',
        )
        parser.add_argument(
            '--threshold',
            type=float,
            default=0.5,
            help='during evaluation, threshold for choosing '
            'ref class; only applies to binary '
            'classification',
        )
        # interactive mode
        parser.add_argument(
            '--print-scores',
            type='bool',
            default=False,
            help='print probability of chosen class during ' 'interactive mode',
        )
        # miscellaneous arguments
        parser.add_argument(
            '--data-parallel',
            type='bool',
            default=False,
            help='uses nn.DataParallel for multi GPU',
        )
        parser.add_argument(
            '--classes-from-file',
            type=str,
            default=None,
            help='loads the list of classes from a file',
        )
        parser.add_argument(
            '--ignore-labels',
            type='bool',
            default=None,
            help='Ignore labels provided to model',
        )
        parser.add_argument(
            '--update-classifier-head-only',
            type='bool',
            default=False,
            help='Freeze the encoder and update the classifier head only',
        )
        parser.set_defaults(use_reply='none')
        return parser

    def __init__(self, opt: Opt, shared=None):
        init_model, self.is_finetune = self._get_init_model(opt, shared)
        super().__init__(opt, shared)

        # set up classes
        if opt.get('classes') is None and opt.get('classes_from_file') is None:
            raise RuntimeError(
                'Must specify --classes or --classes-from-file argument.'
            )
        if not shared:
            if opt['classes_from_file'] is not None:
                with PathManager.open(opt['classes_from_file']) as f:
                    self.class_list = f.read().splitlines()
            else:
                self.class_list = opt['classes']
            self.class_dict = {val: i for i, val in enumerate(self.class_list)}
            if opt.get('class_weights', None) is not None:
                self.class_weights = opt['class_weights']
            else:
                self.class_weights = [1.0 for c in self.class_list]
            self.reset_metrics()
        else:
            self.class_list = shared['class_list']
            self.class_dict = shared['class_dict']
            self.class_weights = shared['class_weights']

        # in binary classfication, opt['threshold'] applies to ref class
        if opt['ref_class'] is None or opt['ref_class'] not in self.class_dict:
            self.ref_class = self.class_list[0]
        else:
            self.ref_class = opt['ref_class']
            ref_class_id = self.class_list.index(self.ref_class)
            if ref_class_id != 0:
                # move to the front of the class list
                self.class_list.insert(0, self.class_list.pop(ref_class_id))

        # set up threshold, only used in binary classification
        if len(self.class_list) == 2 and opt.get('threshold', 0.5) != 0.5:
            self.threshold = opt['threshold']
        else:
            self.threshold = None

        # set up calculating auc
        self.calc_auc = opt.get('area_under_curve_digits', -1) > 0

        if self.calc_auc:
            self.auc_bucket_decimal_size = opt.get('area_under_curve_digits')
            if opt.get('area_under_curve_class') is None:
                # self.auc_class_ind
                interested_classes = self.class_list
            else:
                interested_classes = opt.get('area_under_curve_class')
            try:
                self.auc_class_indices = [
                    self.class_dict[class_name] for class_name in interested_classes
                ]
            except Exception:
                raise RuntimeError(
                    f'The inputted classes for auc were probably invalid.\n Current class names: {self.class_list} \n Names of AUC classes passed in: {interested_classes}'
                )
            self.reset_auc()

        # set up model and optimizers
        states = {}
        if shared:
            self.model = shared['model']
        else:
            self.model = self.build_model()
            # freeze the encoder and update the classifier only
            if opt.get("update_classifier_head_only", False):
                for _param_name, _param_value in self.model.named_parameters():
                    if not _param_name.startswith('additional_linear_layer'):
                        _param_value.requires_grad = False

            self.criterion = self.build_criterion()
            if self.model is None or self.criterion is None:
                raise AttributeError(
                    'build_model() and build_criterion() need to return the model or criterion'
                )
            if init_model:
                logging.info(f'Loading existing model parameters from {init_model}')
                states = self.load(init_model)
            if self.use_cuda:
                if self.model_parallel:
                    ph = PipelineHelper()
                    ph.check_compatibility(self.opt)
                    self.model = ph.make_parallel(self.model)
                else:
                    self.model.cuda()
                if self.data_parallel:
                    self.model = torch.nn.DataParallel(self.model)
                self.criterion.cuda()

            train_params = trainable_parameters(self.model)
            total_params = total_parameters(self.model)
            logging.info(
                f"Total parameters: {total_params:,d} ({train_params:,d} trainable)"
            )

        if shared:
            # We don't use get here because hasattr is used on optimizer later.
            if 'optimizer' in shared:
                self.optimizer = shared['optimizer']
        elif self._should_initialize_optimizer():
            optim_params = [p for p in self.model.parameters() if p.requires_grad]
            self.init_optim(optim_params)
            self.build_lr_scheduler(states, hard_reset=self.is_finetune)

    def build_criterion(self):
        weight_tensor = torch.FloatTensor(self.class_weights)
        return torch.nn.CrossEntropyLoss(weight=weight_tensor, reduction='none')

    def share(self):
        """
        Share model parameters.
        """
        shared = super().share()
        shared['class_dict'] = self.class_dict
        shared['class_list'] = self.class_list
        shared['class_weights'] = self.class_weights
        shared['model'] = self.model
        if hasattr(self, 'optimizer'):
            shared['optimizer'] = self.optimizer
        return shared

    def _get_labels(self, batch):
        """
        Obtain the correct labels.

        Raises a ``KeyError`` if one of the labels is not in the class list.
        """
        try:
            labels_indices_list = [self.class_dict[label] for label in batch.labels]
        except KeyError as e:
            warn_once('One of your labels is not in the class list.')
            raise e

        labels_tensor = torch.LongTensor(labels_indices_list)
        if self.use_cuda:
            labels_tensor = labels_tensor.cuda()
        return labels_tensor

    def _update_confusion_matrix(self, batch, predictions):
        """
        Update the confusion matrix given the batch and predictions.

        :param predictions:
            (list of string of length batchsize) label predicted by the
            classifier
        :param batch:
            a Batch object (defined in torch_agent.py)
        """
        f1_dict = {}
        for class_name in self.class_list:
            prec_str = f'class_{class_name}_prec'
            recall_str = f'class_{class_name}_recall'
            f1_str = f'class_{class_name}_f1'
            precision, recall, f1 = ConfusionMatrixMetric.compute_metrics(
                predictions, batch.labels, class_name
            )
            f1_dict[class_name] = f1
            self.record_local_metric(prec_str, precision)
            self.record_local_metric(recall_str, recall)
            self.record_local_metric(f1_str, f1)
        self.record_local_metric('weighted_f1', WeightedF1Metric.compute_many(f1_dict))

    def _format_interactive_output(self, probs, prediction_id):
        """
        Format interactive mode output with scores.
        """
        preds = []
        for i, pred_id in enumerate(prediction_id.tolist()):
            prob = round_sigfigs(probs[i][pred_id], 4)
            preds.append(
                'Predicted class: {}\nwith probability: {}'.format(
                    self.class_list[pred_id], prob
                )
            )
        return preds

    def _update_aucs(self, batch, probs):
        probs_arr = probs.detach().cpu().numpy()
        for index, curr_auc in zip(self.auc_class_indices, self.aucs):
            class_probs = probs_arr[:, index]
            curr_auc.update_raw(batch.labels, class_probs, self.class_list[index])

    def train_step(self, batch):
        """
        Train on a single batch of examples.
        """
        if batch.text_vec is None:
            return Output()
        self.model.train()
        self.optimizer.zero_grad()

        # calculate loss
        labels = self._get_labels(batch)
        scores = self.score(batch)
        loss = self.criterion(scores, labels)
        self.record_local_metric('loss', AverageMetric.many(loss))
        loss = loss.mean()
        loss.backward()
        self.update_params()

        # get predictions
        _, prediction_id = torch.max(scores.cpu(), 1)
        preds = [self.class_list[idx] for idx in prediction_id]
        self._update_confusion_matrix(batch, preds)

        return Output(preds)

    def eval_step(self, batch):
        """
        Evaluate a single batch of examples.
        """
        if batch.text_vec is None:
            return

        self.model.eval()
        scores = self.score(batch)
        probs = F.softmax(scores, dim=1)

        if self.calc_auc:
            self._update_aucs(batch, probs)

        if self.threshold is None:
            _, prediction_id = torch.max(probs.cpu(), 1)
        else:
            ref_prob = probs.cpu()[:, 0]
            # choose ref class if Prob(ref class) > threshold
            prediction_id = (ref_prob <= self.threshold).to(torch.int64)
        preds = [self.class_list[idx] for idx in prediction_id]
        if batch.labels is None or self.opt['ignore_labels']:
            # interactive mode
            if self.opt.get('print_scores', False):
                preds = self._format_interactive_output(probs, prediction_id)
        else:
            labels = self._get_labels(batch)
            loss = self.criterion(scores, labels)
            self.record_local_metric('loss', AverageMetric.many(loss))
            loss = loss.mean()
            self._update_confusion_matrix(batch, preds)

        if self.opt.get('print_scores', False):
            return Output(preds, class_list=[self.class_list], probs=probs.cpu())
        else:
            return Output(preds)

    def score(self, batch):
        """
        Given a batch and labels, returns the scores.

        :param batch:
            a Batch object (defined in torch_agent.py)
        :return:
            a [bsz, num_classes] FloatTensor containing the score of each
            class.
        """
        raise NotImplementedError('Abstract class: user must implement score()')

    def reset_auc(self):
        if self.calc_auc:
            self.aucs = [
                AUCMetrics(
                    class_name=self.class_list[index],
                    max_bucket_dec_places=self.auc_bucket_decimal_size,
                )
                for index in self.auc_class_indices
            ]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = "glue"


class TestColaTeacher(AutoTeacherTest):
    task = "glue:cola"


class TestMnliTeacher(AutoTeacherTest):
    task = "glue:mnli"


class TestMrpcTeacher(AutoTeacherTest):
    task = "glue:mrpc"


class TestQnliTeacher(AutoTeacherTest):
    task = "glue:qnli"


class TestQqpTeacher(AutoTeacherTest):
    task = "glue:qqp"


class TestRteTeacher(AutoTeacherTest):
    task = "glue:rte"


class TestSst2Teacher(AutoTeacherTest):
    task = "glue:sst2"


class TestStsbTeacher(AutoTeacherTest):
    task = "glue:stsb"


class TestWnliTeacher(AutoTeacherTest):
    task = "glue:wnli"



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BlenderBot2 Agent Code.

BlenderBot 2 combines a long-term memory module with a retriever module.

The Search Query Generator generates a query that tells BB2 to either access
its memory or access the internet.

The Memory Decoder examines the context and generates memories to write to
the long-term memory module.
"""
import torch
import torch.nn
import torch.nn.functional as F
from typing import Union, Dict, List, Tuple, Optional, Any

from parlai.agents.fid.fid import FidAgent
from parlai.agents.rag.args import DPR_ZOO_MODEL, QUERY_MODEL_TYPES
from parlai.agents.rag.rag import RagAgent
from parlai.agents.rag.model_types import (
    RagTurn,
    RagSequence,
    RagToken,
    RagModelInterface,
)
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import Batch
from parlai.tasks.wizard_of_internet.constants import (
    SELECTED_DOCS,
    SELECTED_DOCS_TITLES,
    SELECTED_SENTENCES,
)
from parlai.utils.torch import padded_3d

from .modules import (
    BlenderBot2RagModel,
    T5BlenderBot2RagModel,
    BlenderBot2FidModel,
    T5BlenderBot2FidModel,
)
from .sub_modules import RetrievalType, KnowledgeAccessMethod
from parlai.agents.fid.fid import SearchQuerySearchEngineFiDAgent


ZOO_QUERY_GENERATOR = 'zoo:blenderbot2/query_generator/model'
ZOO_MEMORY_DECODER = 'zoo:blenderbot2/memory_decoder/model'


class BlenderBot2ModelTypeMixin(RagModelInterface):
    """
    Override Normal RAG Model Types, in case we retrieve from both memory and search.
    """

    def __init__(self, opt: Opt, null_idx: int):
        super().__init__(opt, null_idx)
        if (
            KnowledgeAccessMethod(opt['knowledge_access_method'])
            is KnowledgeAccessMethod.ALL
        ):
            self.n_docs *= 2


class BlenderBot2RagSequence(BlenderBot2ModelTypeMixin, RagSequence):
    def augment_batch_for_generation(
        self, batch: Batch, model: BlenderBot2RagModel
    ) -> Batch:
        """
        Augment batch for generation.

        For RAG Sequence, we retrieve prior to generation, as we do not consider the
        document probabilities until after generating all of the beams.

        :param batch:
            batch to augment
        :param model:
            model to possibly help with augmenting

        :return batch:
            return batch with text vec swapped out.
        """
        (expanded_input, _, doc_scores) = model.retrieve_and_concat(
            batch.text_vec,
            batch.text_vec.ne(self.null_idx).sum(1),
            batch.query_generator_vec,
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.memory_vec,
            batch.num_memories,
            batch.gold_doc_vec,
            batch.gold_doc_title_vec,
            batch.num_gold_docs,
            batch.memory_decoder_vec,
            batch.num_memory_decoder_vecs,
        )
        doc_log_probs = F.log_softmax(doc_scores, dim=1)
        batch.src_text_vec = batch.text_vec
        batch.text_vec = expanded_input
        batch.doc_log_probs = doc_log_probs
        batch.batchsize = batch.text_vec.size(0)

        return batch

    def get_generation_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        Optional[torch.LongTensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
    ]:
        """
        For RAG Sequence, we retrieve prior to generation.
        """
        assert batch.text_vec is not None
        return (
            batch.text_vec,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )


class BlenderBot2RagToken(BlenderBot2ModelTypeMixin, RagToken):
    pass


class BlenderBot2RagTurn(BlenderBot2ModelTypeMixin, RagTurn):
    pass


RAG_MODELS = {
    'sequence': BlenderBot2RagSequence,
    'token': BlenderBot2RagToken,
    'turn': BlenderBot2RagTurn,
}


class BlenderBot2RagAgent(RagAgent):
    """
    Subclass RagAgent to provide BlenderBot2Model with appropriate inputs (specifically,
    memory vectors).
    """

    model: BlenderBot2RagModel

    ##########################
    # Housekeeping functions #
    ##########################
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add RAG Args.
        """
        RagAgent.add_cmdline_args(parser, partial_opt)
        SearchQuerySearchEngineFiDAgent.add_cmdline_args(parser, partial_opt)
        bb2_group = parser.add_argument_group('BlenderBot2 Args')
        bb2_group.add_argument(
            '--knowledge-access-method',
            type=str,
            default=KnowledgeAccessMethod.CLASSIFY.value,
            choices=[r.value for r in KnowledgeAccessMethod],
            help='How to access knowledge for BlenderBot2 '
            'classify => classify the input text, determine which knowledge to access\n'
            'memory_only => only access memories\n'
            'search_only => only access search\n'
            'all => for each input, access from memories and search\n'
            'none => do not access any knowledge.\n',
        )
        bb2_group.add_argument(
            '--memory-key',
            type=str,
            default='full_text',
            help='Field in the observation from which to read memories.',
        )
        bb2_group.add_argument(
            '--query-generator-key',
            type=str,
            default='full_text',
            help='Field for input to the knowledge access classifier.',
        )
        bb2_group.add_argument(
            '--gold-document-key',
            type=str,
            default=SELECTED_DOCS,
            help='Field for selected docs.',
        )
        bb2_group.add_argument(
            '--gold-sentence-key',
            type=str,
            default=SELECTED_SENTENCES,
            help='Field for selected sentences',
        )
        bb2_group.add_argument(
            '--gold-document-titles-key',
            type=str,
            default=SELECTED_DOCS_TITLES,
            help='Field for selected docs titles.',
        )
        bb2_group.add_argument(
            '--insert-gold-docs',
            type='bool',
            default=False,
            help='Set true to insert gold docs into retrieved docs.',
        )
        bb2_group.add_argument(
            '--memory-extractor-phrase',
            type=str,
            default='persona:',
            help="phrase used to extract memories from `--memory-key` in the observation. "
            "For example, set to 'your persona:' to limit memories to only lines that "
            "contain 'your persona:'",
        )
        bb2_group.add_argument(
            '--retriever-ignore-phrase',
            type=str,
            default='persona:',
            help='filter input to the global knowledge retriever such that any utterance containing '
            'the phrase will not be given as input.',
        )
        q_gen_group = parser.add_argument_group('BlenderBot2 Query Generator Args')
        q_gen_group.add_argument(
            '--query-generator-ignore-phrase',
            type=str,
            default='persona:',
            help='filter input to the query generator such that any utterance containing '
            'the phrase will not be given as input.',
        )
        q_gen_group.add_argument(
            '--query-generator-model-file',
            type=str,
            default=ZOO_QUERY_GENERATOR,
            help='path to a query generator; specify if searching OR classifying inputs.',
        )
        q_gen_group.add_argument(
            '--query-generator-delimiter',
            type=str,
            default='\n',
            help='delimiter for the query generator',
        )
        q_gen_group.add_argument(
            '--query-generator-inference',
            type=str,
            default='beam',
            help='query generator inference type',
        )
        q_gen_group.add_argument(
            '--query-generator-beam-size', type=int, default=1, help='SQ Gen Beam Size'
        )
        q_gen_group.add_argument(
            '--query-generator-beam-min-length',
            type=int,
            default=2,
            help='SQ Gen Beam Min Length',
        )
        q_gen_group.add_argument(
            '--query-generator-truncate',
            type=int,
            default=-1,
            help='Specify >0 for truncation to SQ generator',
        )
        bb2_group.add_argument(
            '--memory-retriever-truncate',
            type=int,
            default=-1,
            help='Specify >0 for truncation to the memory retriever.',
        )
        bb2_group.add_argument(
            '--retriever-delimiter',
            type=str,
            default='\n',
            help='delimiter for the retriever',
        )
        bb2_group.add_argument(
            '--share-search-and-memory-query-encoder',
            type='bool',
            default=False,
            help='if true, query encoder is shared between search and memory retrievers.',
        )
        bb2_group.add_argument(
            '--memory-reader-model',
            type=str,
            default=None,
            choices=QUERY_MODEL_TYPES,
            help='Model for accessing the memory',
        )
        bb2_group.add_argument(
            '--memory-doc-title-delimiter',
            type=str,
            default=' / ',
            help='title delimiter for memory docs',
        )
        bb2_group.add_argument(
            '--memory-writer-model',
            type=str,
            default='bert',
            hidden=True,
            help='model for writing the memories',
        )
        bb2_group.add_argument(
            '--memory-writer-model-file',
            type=str,
            default=DPR_ZOO_MODEL,
            hidden=True,
            help='model file for memory writer',
        )
        memory_decoder = parser.add_argument_group('BlenderBot2 Memory Decoder Args')
        memory_decoder.add_argument(
            '--memory-decoder-key',
            type=str,
            default='full_text',
            help='key of the observation for the memory decoder',
        )
        memory_decoder.add_argument(
            '--memory-decoder-ignore-phrase',
            type=str,
            default='persona:',
            help='filter input to the memory decoder such that any utterance containing '
            'the phrase will not be given as input.',
        )
        memory_decoder.add_argument(
            '--memory-decoder-model-file',
            type=str,
            default=ZOO_MEMORY_DECODER,
            help='path to a memory decoder.',
        )
        memory_decoder.add_argument(
            '--memory-decoder-delimiter',
            type=str,
            default='\n',
            help='delimiter for the memory decoder',
        )
        memory_decoder.add_argument(
            '--memory-decoder-beam-size',
            type=int,
            default=3,
            help='memory decoder Beam Size',
        )
        memory_decoder.add_argument(
            '--memory-decoder-beam-min-length',
            type=int,
            default=10,
            help='memory decoder Beam Min Length',
        )
        memory_decoder.add_argument(
            '--memory-decoder-truncate',
            type=int,
            default=-1,
            help='Specify >0 for truncation to memory decoder',
        )
        memory_decoder.add_argument(
            '--memory-decoder-one-line-memories',
            type='bool',
            default=False,
            help='specify to combine memories on one line, rather than several.',
        )
        return parser

    @property
    def rag_model_type(self) -> str:
        return self._rag_model_type

    @rag_model_type.setter
    def rag_model_type(self, model: str):
        self._rag_model_type = model
        self._rag_model_interface = RAG_MODELS[model](self.opt, self.NULL_IDX)

    def build_model(self) -> BlenderBot2RagModel:
        """
        Build and return BlenderBot2RagModel.
        """
        if self.generation_model == 't5':
            model = T5BlenderBot2RagModel(self.opt, self.dict)
        else:
            model = BlenderBot2RagModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    @classmethod
    def upgrade_opt(cls, opt_from_disk: Opt):
        # call the parent upgrades
        opt_from_disk = super().upgrade_opt(opt_from_disk)

        if 'memory_doc_delimiter' not in opt_from_disk:
            # 2020-06-22 old delimiter was ':'
            opt_from_disk['memory_doc_delimiter'] = ':'

        return opt_from_disk

    @staticmethod
    def update_state_dict(
        opt: Opt, state_dict: Dict[str, torch.Tensor], model: torch.nn.Module
    ):
        """
        Override RagAgent.update_state_dict to store long term memory state.
        """
        state_dict = RagAgent.update_state_dict(opt, state_dict, model)
        # 1. Retriever state
        if not [k for k in state_dict if 'long_term_memory' in k]:
            long_term_memory_state = {
                f"long_term_memory.{k}": v
                for k, v in model.long_term_memory.state_dict().items()  # type: ignore
            }
            state_dict.update(long_term_memory_state)
        return state_dict

    ###############################
    # Text/Tokenization Overrides #
    ###############################
    def observe(self, observation: Union[Dict, Message]) -> Message:
        """
        Overrides TA.observe to tokenize various additional vectors.
        """
        observation = super().observe(observation)
        if 'memory_vec' not in observation and self.opt['memory_key'] in observation:
            self._set_memory_vec(observation)
        if (
            'query_generator_vec' not in observation
            and self.opt['query_generator_key'] in observation
        ):
            self._set_query_generator_vec(observation)
        if 'gold_doc_vec' not in observation and all(
            k in observation
            for k in [
                self.opt['gold_document_key'],
                self.opt['gold_sentence_key'],
                self.opt['gold_document_titles_key'],
            ]
        ):
            self._set_gold_doc_vec(observation)
        if (
            'memory_decoder_vec' not in observation
            and self.opt['memory_decoder_key'] in observation
        ):
            self._set_memory_decoder_vec(observation)
        return observation

    def _filter_text(self, text: str, filter_phrase: str, delimiter: str = '\n') -> str:
        """
        Filter text such that utterances containing a filter phrase are removed.

        :param text:
            text to filter
        :param filter_phrase:
            phrase on which to filter
        :param delimiter:
            optional extra delimiter on which to split

        :return text:
            return the text after filtering (including or excluding) turns with the filter phrase.
        """
        split_text = [
            t
            for tt in text.split(self.opt.get('delimiter', '\n'))
            for t in tt.split('\n')
        ]
        turns = [t for t in split_text if filter_phrase not in t]
        if not turns:
            new_text = text
        else:
            new_text = delimiter.join(turns)
        return new_text

    def _remove_person_tokens(self, text: str) -> str:
        """
        Remove person tokens from a text input.
        """
        return text.replace(f'{self.P1_TOKEN} ', '').replace(f'{self.P2_TOKEN} ', '')

    def _set_query_vec(self, observation: Message) -> Message:
        """
        Override RAG.set_query_vec to optionally filter phrases.
        """
        query_str = observation[self._query_key]
        if self.opt['retriever_ignore_phrase']:
            query_str = self._filter_text(
                query_str,
                self.opt['retriever_ignore_phrase'],
                delimiter=self.opt['retriever_delimiter'],
            )
        if self.add_person_tokens:
            query_str = self._remove_person_tokens(query_str)
        observation['query_vec'] = self.model.tokenize_query(query_str)
        return observation

    def _set_memory_vec(self, observation: Message) -> Message:
        """
        Tokenize the memories for use in long-term memory scoring.

        :param observation:
            observation with input text.

        :return observation:
            return observation with memory vec.
        """
        mem_vecs = None
        method = KnowledgeAccessMethod(self.opt['knowledge_access_method'])
        if method in [
            KnowledgeAccessMethod.ALL,
            KnowledgeAccessMethod.CLASSIFY,
            KnowledgeAccessMethod.MEMORY_ONLY,
        ]:
            memories = observation[self.opt['memory_key']]
            if isinstance(memories, str):
                memories = [
                    t
                    for tt in memories.split(self.opt.get('delimiter', '\n'))
                    for t in tt.split('\n')
                ]
            assert isinstance(memories, list)
            if self.opt['memory_extractor_phrase']:
                # extract text lines only containing the memory extractor phrase
                memories = [
                    m for m in memories if self.opt['memory_extractor_phrase'] in m
                ]
            if memories:
                mem_vecs = [self.model.tokenize_memory(mem) for mem in memories]

        observation['memory_vec'] = mem_vecs
        return observation

    def _set_query_generator_vec(self, observation: Message) -> Message:
        """
        Tokenize text for use in the query generator.

        :param observation:
            observation with input text.

        :return observation:
            return observation with query generator vec.
        """
        query_generator_vec = None
        method = KnowledgeAccessMethod(self.opt['knowledge_access_method'])
        if (
            method
            in [
                KnowledgeAccessMethod.ALL,
                KnowledgeAccessMethod.CLASSIFY,
                KnowledgeAccessMethod.SEARCH_ONLY,
            ]
            and self.model.has_query_generator()
        ):
            query_generator_input = observation[self.opt['query_generator_key']]
            if self.opt['query_generator_ignore_phrase']:
                query_generator_input = self._filter_text(
                    query_generator_input,
                    self.opt['query_generator_ignore_phrase'],
                    self.opt['query_generator_delimiter'],
                )
            if self.add_person_tokens:
                query_generator_input = self._remove_person_tokens(
                    query_generator_input
                )
            query_generator_vec = self.model.tokenize_query_generator_input(
                query_generator_input
            )

        observation['query_generator_vec'] = query_generator_vec
        return observation

    def _set_gold_doc_vec(self, observation: Message) -> Message:
        """
        Tokenize the gold documents, in case we want to include in retrieved documents.

        We chunk up the docs and try to find the chunk that contains the selected sentence.

        If we can't find it, we just use the first chunk.

        :param observation:
            observation with input text.

        :return observation:
            return observation with gold doc vec.
        """
        if not observation[self.opt['gold_document_key']]:
            return observation
        doc_vecs = None
        doc_title_vecs = None
        method = KnowledgeAccessMethod(self.opt['knowledge_access_method'])
        chunk_len = self.opt.get("splitted_chunk_length", 256)
        if method in [
            KnowledgeAccessMethod.ALL,
            KnowledgeAccessMethod.CLASSIFY,
            KnowledgeAccessMethod.SEARCH_ONLY,
        ]:
            selected_documents = observation[self.opt['gold_document_key']]
            sentences = observation[self.opt['gold_sentence_key']]
            document_titles = observation[self.opt['gold_document_titles_key']]
            if isinstance(selected_documents, str):
                documents = [selected_documents]
            assert isinstance(selected_documents, list)

            documents = []
            for doc in selected_documents:
                # Try to find the chunk with the selected sentence
                used_chunk = None
                words = doc.split(' ')
                chunks = [
                    ' '.join(words[i : i + chunk_len])
                    for i in range(0, len(words), chunk_len)
                ]
                for chunk in chunks:
                    if any(s in chunk for s in sentences):
                        used_chunk = chunk
                        break
                if not used_chunk:
                    used_chunk = chunks[0]
                documents.append(used_chunk)

            if documents:
                doc_vecs = [self.dict.txt2vec(doc) for doc in documents]
                doc_title_vecs = [self.dict.txt2vec(title) for title in document_titles]

        observation['gold_doc_vec'] = doc_vecs
        observation['gold_doc_title_vec'] = doc_title_vecs
        return observation

    def _set_memory_decoder_vec(self, observation: Message) -> Message:
        """
        Tokenize the input to the memory decoder.

        :param observation:
            observation with input text.

        :return observation:
            return observation with memory vec.
        """
        memory_decoder_vec = None
        method = KnowledgeAccessMethod(self.opt['knowledge_access_method'])
        if (
            method
            in [
                KnowledgeAccessMethod.ALL,
                KnowledgeAccessMethod.CLASSIFY,
                KnowledgeAccessMethod.MEMORY_ONLY,
            ]
            and self.model.has_memory_decoder()
        ):
            memory_decoder_input = observation[self.opt['memory_decoder_key']]
            if self.opt['memory_decoder_ignore_phrase']:
                memory_decoder_input = self._filter_text(
                    memory_decoder_input,
                    self.opt['memory_decoder_ignore_phrase'],
                    self.opt['memory_decoder_delimiter'],
                )
            if self.add_person_tokens:
                memory_decoder_input = self._remove_person_tokens(memory_decoder_input)
            conv_lines = [
                t
                for tt in memory_decoder_input.split(self.opt.get('delimiter', '\n'))
                for t in tt.split('\n')
            ]
            memory_decoder_vec = [
                self.model.tokenize_memory_decoder_input(i) for i in conv_lines
            ]

        observation['memory_decoder_vec'] = memory_decoder_vec
        return observation

    def batchify(self, obs_batch: List[Message], sort: bool = False) -> Batch:
        """
        Overrides RagAgent.batchify to add several input vectors.
        """
        batch = super().batchify(obs_batch, sort)
        valid_exs = [ex for ex in obs_batch if self.is_valid(ex)]
        batch.memory_vec = None
        batch.num_memories = None
        batch.query_generator_vec = None
        batch.gold_doc_vec = None
        batch.gold_doc_title_vec = None
        batch.num_gold_docs = None
        batch.memory_decoder_vec = None
        batch.num_memory_decoder_vecs = None
        if any(ex.get('memory_vec') is not None for ex in valid_exs):
            batch = self._set_batch_memory_vec(valid_exs, batch)
        if any(ex.get('query_generator_vec') is not None for ex in valid_exs):
            batch = self._set_batch_query_generator_vec(valid_exs, batch)
        if any(ex.get('gold_doc_vec') is not None for ex in valid_exs):
            batch = self._set_batch_gold_doc_vec(valid_exs, batch)
        if any(ex.get('memory_decoder_vec') is not None for ex in valid_exs):
            batch = self._set_batch_memory_decoder_vec(valid_exs, batch)
        return batch

    def _set_batch_memory_vec(self, valid_exs: List[Message], batch: Batch) -> Batch:
        """
        Set the memory vec for the batch.
        """
        mems = []
        num_mems = []
        for ex in valid_exs:
            if ex.get('memory_vec') is not None:
                ms, _ = self._pad_tensor(ex['memory_vec'])
                mems.append(ms)
                num_mems.append(len(ex['memory_vec']))
            else:
                num_mems.append(0)
        batch.memory_vec = padded_3d(mems)
        batch.num_memories = torch.LongTensor(num_mems)
        return batch

    def _set_batch_query_generator_vec(
        self, valid_exs: List[Message], batch: Batch
    ) -> Batch:
        """
        Set the query generator vec for the batch.
        """
        _q_gens = [ex.get('query_generator_vec', self.EMPTY) for ex in valid_exs]
        q_gen_vecs, _lens = self._pad_tensor(_q_gens)
        batch.query_generator_vec = q_gen_vecs
        return batch

    def _set_batch_gold_doc_vec(self, valid_exs: List[Message], batch: Batch) -> Batch:
        """
        Set the gold docs vecs for the batch.
        """
        docs = []
        titles = []
        num_docs = []
        for ex in valid_exs:
            if ex.get('gold_doc_vec') is not None:
                ds, _ = self._pad_tensor(ex['gold_doc_vec'])
                ts, _ = self._pad_tensor(ex['gold_doc_title_vec'])
                docs.append(ds)
                titles.append(ts)
                num_docs.append(len(ex['gold_doc_vec']))
            else:
                docs.append(self.EMPTY.unsqueeze(0))
                titles.append(self.EMPTY.unsqueeze(0))
                num_docs.append(0)
        batch.gold_doc_vec = padded_3d(docs)
        batch.gold_doc_title_vec = padded_3d(titles)
        batch.num_gold_docs = torch.LongTensor(num_docs)
        return batch

    def _set_batch_memory_decoder_vec(
        self, valid_exs: List[Message], batch: Batch
    ) -> Batch:
        """
        Set the memory decoder vec for the batch.
        """
        memory_dec_toks = []
        num_memory_dec_toks = []
        for ex in valid_exs:
            if ex.get('memory_decoder_vec') is not None:
                p_sum_vecs, _lens = self._pad_tensor(ex['memory_decoder_vec'])
                memory_dec_toks.append(p_sum_vecs)
                num_memory_dec_toks.append(len(ex['memory_decoder_vec']))
            else:
                num_memory_dec_toks.append(0)
        batch.memory_decoder_vec = padded_3d(memory_dec_toks)
        batch.num_memory_decoder_vecs = torch.LongTensor(num_memory_dec_toks)
        return batch

    def eval_step(self, batch):
        output = super().eval_step(batch)
        if output is None or not hasattr(self.model, 'retriever'):
            return output
        if hasattr(self.model.retriever, 'top_docs'):
            output.top_docs = self.model.retriever.top_docs
        if hasattr(self.model.retriever, 'search_queries'):
            output.search_queries = self.model.retriever.search_queries
        return output

    def _model_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
    ]:
        """
        Override RagAgent._model_input to include several more input vectors.

        See BlenderBot2RagModel.encoder for details.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.memory_vec,
            batch.num_memories,
            batch.query_generator_vec,
            batch.gold_doc_vec,
            batch.gold_doc_title_vec,
            batch.num_gold_docs,
            batch.memory_decoder_vec,
            batch.num_memory_decoder_vecs,
        )

    def compute_loss(
        self, batch: Batch, return_output: bool = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, Any]]:
        """
        Override Rag.compute_loss to add some additional metrics.
        """
        loss, output = super().compute_loss(batch, return_output=True)
        assert isinstance(self.model, BlenderBot2RagModel)
        if (
            KnowledgeAccessMethod(self.opt['knowledge_access_method'])
            is KnowledgeAccessMethod.CLASSIFY
            and self.model.has_query_generator()
        ):
            _scores, _preds, enc_state, *_ = output
            _, _, input_turns_cnt, _, _ = enc_state
            retrieval_type = self.model.get_retrieval_type()
            assert isinstance(retrieval_type, torch.Tensor)
            if input_turns_cnt is not None:
                new_ret_type = torch.zeros(input_turns_cnt.size(0))
                offset = 0
                for i in range(input_turns_cnt.size(0)):
                    new_ret_type[i] = retrieval_type[offset]
                    offset += input_turns_cnt[i]
                retrieval_type = new_ret_type
            self.record_local_metric(
                'search_class',
                AverageMetric.many(
                    retrieval_type.eq(RetrievalType.SEARCH.value).int().tolist(),
                    [1] * retrieval_type.size(0),
                ),
            )
            self.record_local_metric(
                'memory_class',
                AverageMetric.many(
                    retrieval_type.eq(RetrievalType.MEMORY.value).int().tolist(),
                    [1] * retrieval_type.size(0),
                ),
            )
            self.record_local_metric(
                'none_class',
                AverageMetric.many(
                    retrieval_type.eq(RetrievalType.NONE.value).int().tolist(),
                    [1] * retrieval_type.size(0),
                ),
            )
        if return_output:
            return loss, output
        else:
            return loss


class BlenderBot2FidAgent(FidAgent, BlenderBot2RagAgent):
    model: BlenderBot2FidModel

    def build_model(self) -> Union[BlenderBot2FidModel, T5BlenderBot2FidModel]:
        if self.generation_model == 't5':
            model = T5BlenderBot2FidModel(self.opt, self.dict)
        else:
            model = BlenderBot2FidModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import abstractmethod
from copy import deepcopy
from collections import defaultdict
from typing import Dict, List, Optional, Tuple
import jsonlines
import os
from parlai.core.message import Message
from tqdm import tqdm
from parlai.core.metrics import F1Metric
from parlai.core.params import ParlaiParser, Opt
from parlai.core.teachers import DialogTeacher
from parlai.utils.data import DatatypeHelper
import parlai.utils.logging as logging
import parlai.tasks.wizard_of_internet.constants as CONST

from .build import build


def get_dtype(opt):
    return DatatypeHelper.fold(opt.get('datatype', 'train'))


def _path(opt):
    build(opt)
    dpath = os.path.join(opt['datapath'], CONST.DATASET_NAME)
    dtype = get_dtype(opt)
    return os.path.join(dpath, f'{dtype}.jsonl')


def get_single_val_from_dict(dialog_json):
    """
    Extracting the single dialogue in the JSON.
    """
    assert len(dialog_json) == 1
    return next(iter(dialog_json.values()))


def parse_agent_message(message_dict, agent_type):
    return {
        CONST.SPEAKER_ID: agent_type,
        CONST.MESSAGE_TEXT: message_dict[CONST.MESSAGE_TEXT],
    }


def parse_apprentice_message(message_dict):
    return parse_agent_message(message_dict, CONST.APPRENTICE)


def wizard_message_has_selection(message):
    sel_docs = message[CONST.CONTEXT][CONST.SELECTED_CONTENTS]
    if len(sel_docs) < 2:
        return False
    return not sel_docs[0][0]


def parse_wizard_message(message_dict, doc_lines_delim):
    def get_knowledge(msg_d):
        knowledge = {
            CONST.RETRIEVED_DOCS: [],
            CONST.SELECTED_DOCS: [],
            CONST.SELECTED_DOCS_TITLES: [],
            CONST.SELECTED_SENTENCES: [],
            CONST.RETRIEVED_DOCS_URLS: [],
            CONST.RETRIEVED_DOCS_TITLES: [],
        }
        docs = msg_d[CONST.CONTEXT][CONST.CONTENTS]
        selections = msg_d[CONST.CONTEXT][CONST.SELECTED_CONTENTS]

        # Checking the option that agents choose if there is no selected sentence
        has_selection = wizard_message_has_selection(msg_d)

        for doc_ind, doc in enumerate(docs):
            doc_lines = []
            doc_lines_selection = selections[doc_ind + 1]
            doc_selected = False
            for line_ind, line in enumerate(doc['content']):
                doc_lines.append(line)
                if has_selection and doc_lines_selection[line_ind]:
                    doc_selected = True
                    knowledge[CONST.SELECTED_SENTENCES].append(line)
            full_doc = doc_lines_delim.join(doc_lines)
            knowledge[CONST.RETRIEVED_DOCS].append(full_doc)
            knowledge[CONST.RETRIEVED_DOCS_TITLES].append(doc['title'])
            knowledge[CONST.RETRIEVED_DOCS_URLS].append(doc['url'])
            if doc_selected:
                knowledge[CONST.SELECTED_DOCS_TITLES].append(doc['title'])
                knowledge[CONST.SELECTED_DOCS].append(full_doc)

        if not knowledge[CONST.RETRIEVED_DOCS]:
            knowledge[CONST.RETRIEVED_DOCS] = [CONST.NO_RETRIEVED_DOCS_TOKEN]
            knowledge[CONST.RETRIEVED_DOCS_URLS] = [CONST.NO_URLS]

        if not knowledge[CONST.SELECTED_DOCS]:
            knowledge[CONST.SELECTED_DOCS] = [CONST.NO_SELECTED_DOCS_TOKEN]
            knowledge[CONST.SELECTED_SENTENCES] = [CONST.NO_SELECTED_SENTENCES_TOKEN]

        return knowledge

    d = parse_agent_message(message_dict, CONST.WIZARD)
    if message_dict[CONST.ACTION] == CONST.ACTION_WIZARD_TO_APPRENTICE:
        d.update(get_knowledge(message_dict))
    return d


def parse_search_results(message_dict, delim='; '):
    d = {CONST.SPEAKER_ID: CONST.SEARCH_AGENT}
    d[CONST.SEARCH_RESULTS] = message_dict[CONST.CONTEXT][CONST.CONTENTS]
    all_title = [
        f'({i+1}) {doc["title"]}' for i, doc in enumerate(d[CONST.SEARCH_RESULTS])
    ]
    d[CONST.MESSAGE_TEXT] = delim.join(all_title)
    return d


class WizardOfInternetBaseTeacher(DialogTeacher):
    """
    Base Teacher for Wizard of Internet tasks.

    This teachers loads full conversations and all the actions that happens
    during a full dialogue. Teachers that are drived from this class are
    responsible for slicing data and selecting the actions that they need.
    NOTE: Do NOT use this directly, use its children.
    """

    def __init__(self, opt: Opt, shared=None):
        opt = deepcopy(opt)
        self.datatype = get_dtype(opt)
        opt['datafile'] = _path(opt)
        self.include_persona = opt.get('include_persona', CONST.INCLUDE_PERSONA_DEFAULT)
        self.skip_empty_text = opt.get(
            'skip_empty_text', CONST.SKIP_ON_EMPTY_TEXT_DEFAULT
        )
        self.text_flatten_delimeter = opt.get('delimiter', '\n')
        self.docs_delim = opt.get('docs_delimiter', '\n')
        self.docs_titles_delimeter = opt.get('docs_title_delimiter', '\n')
        self.doc_lines_delim = opt.get('doc_lines_delimiter', '\n')
        self.id = 'WizInternetBase'
        super().__init__(opt, shared=shared)

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group('Wizard Base Dialog Teacher arguments')
        arg_group.add_argument(
            '--include-persona',
            type='bool',
            default=CONST.INCLUDE_PERSONA_DEFAULT,
            help='Whether to include the apprentice persona in text',
        )
        arg_group.add_argument(
            '--skip-empty-text',
            type='bool',
            default=CONST.SKIP_ON_EMPTY_TEXT_DEFAULT,
            help='Whether to skip response to empty messages (may happen if persona is not included)',
        )
        return parser

    def _load_data(self, datafile):
        logging.info(f'Loading data from {datafile} ...')
        dialogs = []
        with jsonlines.open(datafile, 'r') as fin:
            for dialog_json in tqdm(fin):
                dialogs.append(self._get_episode_examples(dialog_json))
        return dialogs

    def _get_episode_examples(self, dialog_json):
        data = get_single_val_from_dict(dialog_json)
        persona = data[CONST.PERSONA]
        output = defaultdict(list)
        msg_history = data[CONST.DIALOG_HIST]
        for msg_ind, message in enumerate(msg_history):
            d = {CONST.PERSONA: persona, CONST.TOTAL_CONVERSATION_INDEX: msg_ind}
            action = message[CONST.ACTION]

            # Separating the actions
            if action == CONST.ACTION_APPRENTICE_TO_WIZARD:
                d.update(parse_apprentice_message(message))
            elif action == CONST.ACTION_WIZARD_TO_APPRENTICE:
                # TODO: must avoid having these in the dataset in the first place
                assert message[CONST.MESSAGE_TEXT], 'Empty message text'

                if (  # Getting the search query that Wizard used for this utterance, if any
                    msg_ind > 1
                    and msg_history[msg_ind - 2][CONST.ACTION]
                    == CONST.ACTION_WIZARD_TO_SEARCH_AGENT
                ):
                    d[CONST.SEARCH_QUERY] = msg_history[msg_ind - 2][CONST.MESSAGE_TEXT]
                    # The last search query was the last one before sending the response
                    output[CONST.ACTION_WIZARD_TO_SEARCH_AGENT][-1][1][
                        CONST.IS_LAST_SEARCH_QUERY
                    ] = True
                else:
                    d[CONST.SEARCH_QUERY] = CONST.NO_SEARCH_QUERY_USED

                d.update(
                    parse_wizard_message(message, doc_lines_delim=self.doc_lines_delim)
                )
                if wizard_message_has_selection(message):
                    # Wizard had selection for this utterance, thus its a knowledge piece
                    output[CONST.ACTION_WIZARD_DOC_SELECTION].append((msg_ind, d))
            elif action == CONST.ACTION_WIZARD_TO_SEARCH_AGENT:
                d.update(
                    parse_wizard_message(message, doc_lines_delim=self.doc_lines_delim)
                )
                d[CONST.IS_SEARCH_QUERY] = True
            elif action == CONST.ACTION_SEARCH_AGENT_TO_WIZARD:
                # TODO: remove assert in the final version
                assert (
                    msg_history[msg_ind - 1][CONST.ACTION]
                    == CONST.ACTION_WIZARD_TO_SEARCH_AGENT
                )

                d.update(parse_search_results(message, self.docs_titles_delimeter))
                # Getting the send text (query) from the latest Wizard search
                d[CONST.SEARCH_QUERY] = output[CONST.ACTION_WIZARD_TO_SEARCH_AGENT][-1][
                    1
                ][CONST.MESSAGE_TEXT]

            # TODO: remove on the final version
            assert 'id' in d, str(message)

            # Getting current actions's previous message/action
            if (
                action == CONST.ACTION_APPRENTICE_TO_WIZARD
                and output[CONST.ACTION_WIZARD_TO_APPRENTICE]
            ):
                d[CONST.PARTNER_PREVIOUS_MESSAGE] = output[
                    CONST.ACTION_WIZARD_TO_APPRENTICE
                ][-1]
            elif output[CONST.ACTION_APPRENTICE_TO_WIZARD]:
                d[CONST.PARTNER_PREVIOUS_MESSAGE] = output[
                    CONST.ACTION_APPRENTICE_TO_WIZARD
                ][-1]

            output[action].append((msg_ind, d))
            output[CONST.ACTION_ALL].append(d)

        return output

    def create_parlai_message(self, dict_message: Dict):
        parlai_msg = Message(
            {
                CONST.SPEAKER_ID: dict_message[CONST.SPEAKER_ID],
                CONST.LABELS: [dict_message[CONST.MESSAGE_TEXT]],
            }
        )
        prv_msg = dict_message.get(CONST.PARTNER_PREVIOUS_MESSAGE)
        if prv_msg:
            parlai_msg[CONST.MESSAGE_TEXT] = prv_msg[1][CONST.MESSAGE_TEXT]
        else:
            parlai_msg[CONST.MESSAGE_TEXT] = ''
        return parlai_msg

    @abstractmethod
    def _teacher_action_type(self) -> str:
        """
        Is this for a Wizard or Apprentice Dialogue agent.
        """

    def additional_message_content(self, parlai_message: Message, action: Dict):
        """
        Children of this class may override this method to add extra content to message.

        It adds components from the original `action` (which is a regular dict) to the
        ParlAI message object `parlai_message`
        """
        pass

    def _opening_message_text(self, parlai_message: Message, action: Dict):
        """
        Handles the first message if this agent is has the opening message.
        """
        if not self.include_persona:
            return

        persona = action[CONST.PERSONA]
        curr_text = parlai_message[CONST.MESSAGE_TEXT]
        if curr_text:
            new_text = f'{persona}{self.text_flatten_delimeter}{curr_text}'
        else:
            new_text = persona

        parlai_message.force_set(CONST.MESSAGE_TEXT, new_text)

    def _should_skip(self, message: Message):
        if not self.skip_empty_text:
            return False
        return not message[CONST.MESSAGE_TEXT].strip()

    def setup_data(self, datafile) -> Message:
        for message, episode_started in self.teacher_setup_data(datafile):
            if not self._should_skip(message):
                yield message, episode_started

    def teacher_setup_data(self, datafile) -> Message:
        for data in self._load_data(datafile):
            started = True
            for idx, (_, act) in enumerate(data[self._teacher_action_type()]):
                parlai_msg = self.create_parlai_message(act)
                if idx == 0 and self.include_persona:
                    self._opening_message_text(parlai_msg, act)
                self.additional_message_content(parlai_msg, act)
                yield parlai_msg, started
                started = False


###############################################################
#                                                             #
# Dialog Teachers                                             #
#                                                             #
###############################################################


class ApprenticeDialogTeacher(WizardOfInternetBaseTeacher):
    def __init__(self, opt, shared=None):
        super().__init__(opt, shared=shared)
        self.id = 'WizInternetApprenticeTeacher'

    def _teacher_action_type(self) -> str:
        return CONST.ACTION_APPRENTICE_TO_WIZARD


class WizardDialogTeacher(WizardOfInternetBaseTeacher):
    def __init__(self, opt, shared=None):
        self.prepend_gold_knowledge = opt.get('prepend_gold_knowledge')
        self.gold_knowledge_delimiter = opt.get('gold_knowledge_delimiter', '\n')
        super().__init__(opt, shared=shared)
        self.id = 'WizInternetWizardTeacher'

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group('Wizard Dialog Knowledge arguments')
        arg_group.add_argument(
            '--prepend-gold-knowledge',
            type='bool',
            default=False,
            help='If true, prepend text with checked sentences',
        )
        return parser

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ) -> None:
        if (
            (
                teacher_action[CONST.SELECTED_SENTENCES][0]
                == CONST.NO_SELECTED_SENTENCES_TOKEN
            )
            or (model_response.is_padding())
            or ('text' not in model_response)
        ):
            # Has NOT selected knowledge or a is batch padding message
            return

        resp = model_response['text']
        self.metrics.add(
            'knowledge_f1_docs',
            F1Metric.compute(resp, [' '.join(teacher_action[CONST.SELECTED_DOCS])]),
        )
        self.metrics.add(
            'knowledge_f1_max_docs', F1Metric.compute(resp, CONST.SELECTED_DOCS)
        )
        self.metrics.add(
            'knowledge_f1_sentences',
            F1Metric.compute(
                resp, [' '.join(teacher_action[CONST.SELECTED_SENTENCES])]
            ),
        )
        self.metrics.add(
            'knowledge_f1_max_sentences',
            F1Metric.compute(resp, CONST.SELECTED_SENTENCES),
        )

    def _teacher_action_type(self) -> str:
        return CONST.ACTION_WIZARD_TO_APPRENTICE

    def additional_message_content(self, parlai_message: Message, action: Dict):
        for item_key in (
            CONST.RETRIEVED_DOCS,
            CONST.RETRIEVED_DOCS_URLS,
            CONST.RETRIEVED_DOCS_TITLES,
            CONST.SELECTED_DOCS,
            CONST.SELECTED_SENTENCES,
            CONST.SELECTED_DOCS_TITLES,
            CONST.SEARCH_QUERY,
        ):
            parlai_message[item_key] = action[item_key]

    def teacher_setup_data(self, datafile) -> Message:
        for message, episode_started in super().teacher_setup_data(datafile):
            if self.prepend_gold_knowledge:
                text = message[CONST.MESSAGE_TEXT]
                gold_knowledge = self.gold_knowledge_delimiter.join(
                    message[CONST.SELECTED_SENTENCES]
                )
                message.force_set(
                    CONST.MESSAGE_TEXT,
                    (
                        f'{CONST.KNOWLEDGE_TOKEN} {gold_knowledge} {CONST.END_KNOWLEDGE_TOKEN}'
                        f' {self.gold_knowledge_delimiter} {text}'
                    ),
                )
            yield message, episode_started


class WizardDialogGoldKnowledgeTeacher(WizardDialogTeacher):
    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser.set_params(prepend_gold_knowledge=True)
        return parser


class DefaultTeacher(WizardDialogTeacher):
    pass


###############################################################
#                                                             #
# Search and Knowledge Teachers                               #
#                                                             #
###############################################################


class BaseSQKnowledgeTeacher(WizardOfInternetBaseTeacher):
    """
    Parent class for knowledge and search query generation teachers.

    We need this teacher because of the actions related to these teacher, that is search
    and knowledge selection, happens as a side track to the main conversation.
    Therefore, we do not want to include the history of the messages emitted by these
    agents in the conversatin history.

    Note: this is an abstract class and is not intended for direct use in a task.
    """

    def __init__(self, opt, shared=None):
        self.dialog_history = opt.get('dialog_history', CONST.DIALOG_HIST_DEFAULT)
        super().__init__(opt, shared=shared)
        self.id = 'BaseKnowledgeTeacher'

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group(
            'Base Search Query and Knowledge arguments'
        )
        arg_group.add_argument(
            '--dialog-history',
            type=str,
            choices=[CONST.HISTORY_TYPE.FULL, CONST.HISTORY_TYPE.ONLY_LAST],
            default=CONST.DIALOG_HIST_DEFAULT,
            help='Full dialogue history or the only the last previous message',
        )
        return parser

    def get_message_history(self, dialog_data: Dict, curr_idx: int) -> List[str]:
        message_hist = []
        for act in dialog_data[CONST.ACTION_ALL]:
            if act[CONST.SPEAKER_ID] in (
                CONST.WIZARD,
                CONST.APPRENTICE,
            ) and not act.get(CONST.IS_SEARCH_QUERY, False):
                if act[CONST.TOTAL_CONVERSATION_INDEX] > curr_idx:
                    break
                message_hist.append(act[CONST.MESSAGE_TEXT])

        if self.dialog_history == CONST.HISTORY_TYPE.ONLY_LAST:
            message_hist = [message_hist[-1]]

        return self.text_flatten_delimeter.join(message_hist)

    def teacher_setup_data(self, datafile) -> Message:
        for data in self._load_data(datafile):
            for idx, act in data[self._teacher_action_type()]:
                parlai_msg = self.create_parlai_message(act)
                if self.dialog_history == CONST.HISTORY_TYPE.FULL:
                    parlai_msg.force_set(
                        CONST.MESSAGE_TEXT, self.get_message_history(data, idx)
                    )
                self._opening_message_text(parlai_msg, act)
                self.additional_message_content(parlai_msg, act)
                yield parlai_msg, True


class SearchQueryTeacher(BaseSQKnowledgeTeacher):
    def __init__(self, opt, shared=None):
        self.only_last_search_query = opt.get(
            'only_last_search_query', CONST.ONLY_LAST_QUERY_DEFAULT
        )
        super().__init__(opt, shared=shared)
        self.id = 'SearchQueryGenerationTeacher'

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group('Search Query Teacher')
        arg_group.add_argument(
            '--only-last-search-query',
            type='bool',
            default=CONST.ONLY_LAST_QUERY_DEFAULT,
            help='Whether to include only the last search before sending the response.',
        )
        return parser

    def _teacher_action_type(self) -> str:
        return CONST.ACTION_WIZARD_TO_SEARCH_AGENT

    def additional_message_content(self, parlai_message: Message, action: Dict):
        parlai_message[CONST.IS_LAST_SEARCH_QUERY] = action.get(
            CONST.IS_LAST_SEARCH_QUERY, False
        )

    def teacher_setup_data(self, datafile) -> Message:
        for message, _ in super().teacher_setup_data(datafile):
            if self.only_last_search_query and not message[CONST.IS_LAST_SEARCH_QUERY]:
                continue
            yield message, True


class BaseKnowledgeTeacher(BaseSQKnowledgeTeacher):
    def __init__(self, opt, shared=None):
        super().__init__(opt, shared=shared)
        self.id = 'KnowledgeGenerationTeacher'

    def _teacher_action_type(self) -> str:
        return CONST.ACTION_WIZARD_DOC_SELECTION

    @abstractmethod
    def _knowledge_piece(self):
        """
        Determines the pieces of knowledge (selected content) to retrieve.

        This may be the enitre document, selected sentences or document titles.
        """

    def additional_message_content(self, parlai_message: Message, action: Dict):
        for item_key in (
            CONST.SELECTED_DOCS,
            CONST.SELECTED_DOCS_TITLES,
            CONST.SELECTED_SENTENCES,
        ):
            parlai_message[item_key] = action[item_key]

    def teacher_setup_data(self, datafile) -> Message:
        for message, _ in super().teacher_setup_data(datafile):
            message.force_set(CONST.LABELS, message[self._knowledge_piece()])
            yield message, True


class GoldKnowledgeTeacher(BaseKnowledgeTeacher):
    def _knowledge_piece(self):
        return CONST.SELECTED_SENTENCES


class GoldDocsTeacher(BaseKnowledgeTeacher):
    def _knowledge_piece(self):
        return CONST.SELECTED_DOCS


class GoldDocTitlesTeacher(BaseKnowledgeTeacher):
    def _knowledge_piece(self):
        return CONST.SELECTED_DOCS_TITLES



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
RAG Args.

An area to maintain RAG-Related Arguments
"""
from enum import Enum
from parlai.core.params import ParlaiParser

QUERY_MODEL_TYPES = ['bert', 'bert_from_parlai_rag', 'dropout_poly']
PRETRAINED_RANKER_TYPES = ['wikito', 'reddit']
SMALL_INDEX_TYPES = [None, 'none', 'exact', 'compressed']

DPR_ZOO_MODEL = 'zoo:hallucination/multiset_dpr/hf_bert_base.cp'
TFIDF_ZOO_MODEL = 'zoo:wikipedia_full/tfidf_retriever/model'
WIKIPEDIA_ZOO_PASSAGES = 'zoo:hallucination/wiki_passages/psgs_w100.tsv'
WIKIPEDIA_COMPRESSED_INDEX = 'zoo:hallucination/wiki_index_compressed/compressed_pq'
WIKIPEDIA_EXACT_INDEX = 'zoo:hallucination/wiki_index_exact/exact'

WOW_PASSAGES_PATH = 'zoo:hallucination/wow_passages/wow_articles.paragraphs.tsv'
WOW_INDEX_PATH = 'zoo:hallucination/wow_passages/exact'
WOW_COMPRESSED_INDEX_PATH = 'zoo:hallucination/wow_passages/compressed'

POLYFAISS_ZOO_MODEL = 'zoo:hallucination/dropout_poly/model'
RAG_TOKEN_ZOO_MODEL = 'zoo:hallucination/bart_rag_token/model'
RAG_SEQUENCE_ZOO_MODEL = 'zoo:hallucination/bart_rag_sequence/model'
RAG_TURN_DTT_ZOO_MODEL = 'zoo:hallucination/bart_rag_turn_dtt/model'
RAG_TURN_DO_ZOO_MODEL = 'zoo:hallucination/bart_rag_turn_do/model'
RAG_DPR_POLY_ZOO_MODEL = 'zoo:hallucination/bart_rag_dpr_poly/model'
FID_DPR_ZOO_MODEL = 'zoo:hallucination/bart_fid_dpr/model'
FID_RAG_ZOO_MODEL = 'zoo:hallucination/bart_fid_rag/model'
FID_RAG_DPR_POLY_ZOO_MODEL = 'zoo:hallucination/bart_fid_rag_dpr_poly/model'

# the following args are copied from
# zoo:pretrained_transformers/poly_model_huge_wikito/model.opt
TRANSFORMER_RANKER_BASE_OPT = {
    'text_truncate': 360,
    'label_truncate': 360,
    'history_size': 20,
    'variant': 'xlm',
    'n_layers': 12,
    'n_heads': 12,
    'n_positions': 1024,
    'embedding_size': 768,
    'ffn_size': 3072,
    'activation': 'gelu',
    'attention_dropout': 0.1,
    'relu_dropout': 0.0,
    'dropout': 0.1,
    'n_segments': 2,
    'learn_embeddings': True,
    'dict_endtoken': '__start__',
    'reduction_type': 'mean',
    'dict_tokenizer': 'bpe',
    'dict_lower': True,
    'output_scaling': 0.06,
    'share_encoders': False,
    'learn_positional_embeddings': True,
    'embeddings_scale': False,
}


# Options for poly-encoders
POLYENCODER_OPT_KEYS = [
    'polyencoder_type',
    'poly_n_codes',
    'poly_attention_type',
    'poly_n_codes',
    'poly_attention_type',
    'poly_attention_num_heads',
    'codes_attention_type',
    'codes_attention_num_heads',
]


class RetrieverType(Enum):
    """
    Rag Retriever Types.
    """

    DPR = 'dpr'
    TFIDF = 'tfidf'
    DPR_THEN_POLY = 'dpr_then_poly'
    POLY_FAISS = 'poly_faiss'
    SEARCH_ENGINE = 'search_engine'
    SEARCH_TERM_FAISS = 'search_term_faiss'


def setup_rag_args(parser: ParlaiParser) -> ParlaiParser:
    group = parser.add_argument_group('RAG Model Args')
    # Standard RAG Agent Arguments
    group.add_argument(
        '--generation-model',
        type=str,
        default='bart',
        help='which generation model to use',
        choices=['transformer/generator', 'bart', 't5'],
    )
    group.add_argument(
        '--query-model',
        type=str,
        default='bert',
        help='Which query model to use for DPR.',
        choices=QUERY_MODEL_TYPES,
    )
    group.add_argument(
        '--rag-model-type',
        type=str,
        default='token',
        help='which rag model decoding to use.',
        choices=['token', 'sequence', 'turn'],
    )
    group.add_argument(
        '--thorough',
        type='bool',
        default=False,
        help='whether to use thorough decoding for rag sequence. ',
    )
    modified_group = parser.add_argument_group('Modified RAG Args')
    modified_group.add_argument(
        '--n-extra-positions',
        type=int,
        default=0,
        help='Specify > 0 to include extra positions in the encoder, in which '
        'retrieved knowledge will go. In this setup, knowledge is _appended_ '
        'instead of prepended.',
    )
    modified_group.add_argument(
        '--gold-knowledge-passage-key',
        type=str,
        default='checked_sentence',
        help='key in the observation dict that indicates the gold knowledge passage. '
        'Specify, along with --debug, to compute passage retrieval metrics at train/test time.',
    )
    modified_group.add_argument(
        '--gold-knowledge-title-key',
        type=str,
        default='title',
        help='key in the observation dict that indicates the gold knowledge passage title. '
        'Specify, along with --debug, to compute passage retrieval metrics at train/test time.',
    )
    retriever_group = parser.add_argument_group('RAG Retriever Args')
    retriever_group.add_argument(
        '--rag-retriever-query',
        type=str,
        default='full_history',
        choices=['one_turn', 'full_history'],
        help='What to use as the query for retrieval. `one_turn` retrieves only on the last turn '
        'of dialogue; `full_history` retrieves based on the full dialogue history.',
    )
    retriever_group.add_argument(
        '--rag-retriever-type',
        type=str,
        default=RetrieverType.DPR.value,
        choices=[r.value for r in RetrieverType],
        help='Which retriever to use',
    )
    retriever_group.add_argument(
        '--retriever-debug-index',
        type=str,
        default=None,
        choices=SMALL_INDEX_TYPES,
        help='Load specified small index, for debugging.',
    )
    retriever_group.add_argument(
        '--n-docs', type=int, default=5, help='How many documents to retrieve'
    )
    retriever_group.add_argument(
        '--min-doc-token-length',
        type=int,
        default=64,
        help='minimum amount of information to retain from document. '
        'Useful to define if encoder does not use a lot of BPE token context.',
    )
    retriever_group.add_argument(
        '--max-doc-token-length',
        type=int,
        default=256,
        help='maximum amount of information to retain from document. ',
    )
    retriever_group.add_argument(
        '--rag-query-truncate',
        type=int,
        default=512,
        help='Max token length of query for retrieval.',
    )
    retriever_group.add_argument(
        '--print-docs',
        type='bool',
        default=False,
        help='Whether to print docs; usually useful during interactive mode.',
    )
    dense_retriever_group = parser.add_argument_group(
        'RAG Dense Passage Retriever Args'
    )
    dense_retriever_group.add_argument(
        '--path-to-index',
        type=str,
        default=WIKIPEDIA_COMPRESSED_INDEX,
        help='path to FAISS Index.',
    )
    dense_retriever_group.add_argument(
        '--path-to-dense-embeddings',
        type=str,
        default=None,
        help='path to dense embeddings directory used to build index. '
        'Default None will assume embeddings and index are in the same directory.',
    )
    dense_retriever_group.add_argument(
        '--dpr-model-file', type=str, default=DPR_ZOO_MODEL, help='path to DPR Model.'
    )
    dense_retriever_group.add_argument(
        '--path-to-dpr-passages',
        type=str,
        default=WIKIPEDIA_ZOO_PASSAGES,
        help='Path to DPR passages, used to build index.',
    )
    dense_retriever_group.add_argument(
        '--retriever-embedding-size',
        type=int,
        default=768,
        help='Embedding size of dense retriever',
    )
    tfidf_retriever_group = parser.add_argument_group('RAG TFIDF Retriever Args')
    tfidf_retriever_group.add_argument(
        '--tfidf-max-doc-paragraphs',
        type=int,
        default=-1,
        help='If > 0, limit documents to this many paragraphs',
    )
    tfidf_retriever_group.add_argument(
        '--tfidf-model-path',
        type=str,
        default=TFIDF_ZOO_MODEL,
        help='Optionally override TFIDF model.',
    )
    dpr_poly_retriever_group = parser.add_argument_group('RAG DPR-POLY Retriever Args')
    dpr_poly_retriever_group.add_argument(
        '--dpr-num-docs',
        type=int,
        default=25,
        help='In two stage retrieval, how many DPR documents to retrieve',
    )
    dpr_poly_retriever_group.add_argument(
        '--poly-score-initial-lambda',
        type=float,
        default=0.5,
        help='In two stage retrieval, how much weight to give to the poly scores. '
        'Note: Learned parameter. Specify initial value here',
    )
    dpr_poly_retriever_group.add_argument(
        '--polyencoder-init-model',
        type=str,
        default='wikito',
        help='Which init model to initialize polyencoder with. Specify wikito or reddit to use '
        'models from the ParlAI zoo; otherwise, provide a path to a trained polyencoder',
    )
    poly_faiss_group = parser.add_argument_group('RAG PolyFAISS retriever args')
    poly_faiss_group.add_argument(
        '--poly-faiss-model-file',
        type=str,
        default=None,
        help='path to poly-encoder for use in poly-faiss retrieval.',
    )
    regret_group = parser.add_argument_group("RAG ReGReT args")
    regret_group.add_argument(
        '--regret',
        type='bool',
        default=False,
        help='Retrieve, Generate, Retrieve, Tune. '
        'Retrieve, generate, then retrieve again, and finally tune (refine).',
    )
    regret_group.add_argument(
        '--regret-intermediate-maxlen',
        type=int,
        default=32,
        help='Maximum length in intermediate regret generation',
    )
    regret_group.add_argument(
        '--regret-model-file',
        type=str,
        default=None,
        help='Path to model for initial round of retrieval. ',
    )
    regret_group.add_argument(
        '--regret-dict-file',
        type=str,
        default=None,
        help='Path to dict file for model for initial round of retrieval. ',
    )
    regret_group.add_argument(
        '--regret-override-index',
        type='bool',
        default=False,
        help='Overrides the index used with the ReGReT model, if using separate models. '
        'I.e., the initial round of retrieval uses the same index as specified for the '
        'second round of retrieval',
    )
    indexer_group = parser.add_argument_group("RAG Indexer Args")
    indexer_group.add_argument(
        '--indexer-type',
        type=str,
        default='compressed',
        choices=['exact', 'compressed'],
        help='Granularity of RAG Indexer. Choose compressed to save on RAM costs, at the '
        'possible expense of accuracy.',
    )
    indexer_group.add_argument(
        '--indexer-buffer-size',
        type=int,
        default=65536,
        help='buffer size for adding vectors to the index',
    )
    indexer_group.add_argument(
        '--compressed-indexer-factory',
        type=str,
        default='IVF4096_HNSW128,PQ128',
        help='If specified, builds compressed indexer from a FAISS Index Factory. '
        'see https://github.com/facebookresearch/faiss/wiki/The-index-factory for details',
    )
    indexer_group.add_argument(
        '--compressed-indexer-gpu-train',
        type='bool',
        default=False,
        hidden=True,
        help='Set False to not train compressed indexer on the gpu.',
    )
    indexer_group.add_argument(
        '--compressed-indexer-nprobe',
        type=int,
        default=64,
        help='How many centroids to search in compressed indexer. See '
        'https://github.com/facebookresearch/faiss/wiki/Faiss-indexes#cell-probe-methods-indexivf-indexes '
        'for details',
    )
    # See https://github.com/facebookresearch/faiss/wiki/Faiss-indexes#indexhnsw-variants for details
    indexer_group.add_argument(
        '--hnsw-indexer-store-n',
        type=int,
        default=128,
        hidden=True,
        help='Granularity of  DenseHNSWIndexer. Higher == more accurate, more RAM',
    )
    indexer_group.add_argument(
        '--hnsw-ef-search',
        type=int,
        default=128,
        hidden=True,
        help='Depth of exploration of search for HNSW.',
    )
    indexer_group.add_argument(
        '--hnsw-ef-construction',
        type=int,
        default=200,
        hidden=True,
        help='Depth of exploration at add time for HNSW',
    )
    return parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass, field
from mephisto.operations.registry import register_mephisto_abstraction
from mephisto.abstractions.blueprints.parlai_chat.parlai_chat_blueprint import (
    ParlAIChatBlueprintArgs,
    ParlAIChatBlueprint,
)
from mephisto.abstractions.blueprint import SharedTaskState
from omegaconf import DictConfig, MISSING

WIZARD_INTERNET_PARLAICHAT_BLUEPRINT = 'wizard_internet_parlaichat_blueprint'


@dataclass
class WizardOfInternetBlueprintArgs(ParlAIChatBlueprintArgs):
    _blueprint_type: str = WIZARD_INTERNET_PARLAICHAT_BLUEPRINT
    _group: str = field(
        default='WizardInternetParlAIChatBlueprint',
        metadata={
            'help': """
                ParlAI chat between two agents with one agent having access to a search API
                that retrieves data from internet (common crawl snapshot). In order to run,
                the search API needs to be up and running.
            """
        },
    )

    role_qualification: str = field(
        default=MISSING,
        metadata={
            'help': """
                Specify the role (wizard or apprentice) that agents are trained on
                during their onboarding.
            """
        },
    )

    min_turns: int = field(
        default=MISSING,
        metadata={
            'help': """
                The minimum number of turns before showing the finish button on chat interface
                and allowing the agents to end the conversations cleanly.
            """
        },
    )

    wizard_time_out: int = field(
        default=180,
        metadata={'help': 'Maximum allowed time (seconds) for Wizard, each round.'},
    )

    apprentice_time_out: int = field(
        default=60,
        metadata={'help': 'Maximum allowed time (seconds) for Apprentice, each round.'},
    )

    personas_file: str = field(
        default=MISSING,
        metadata={'help': 'Path to a text file that keeps a list of curated personas.'},
    )

    persona_counts_file: str = field(
        default=MISSING,
        metadata={
            'help': 'A semicolon seperated list of personas and their count (file)'
        },
    )

    shuffle_persona: bool = field(
        default=True, metadata={'help': 'Whether to shuffle the persona list'}
    )

    use_personas_with_replacement: bool = field(
        default=False,
        metadata={'help': 'Using true does not discard personas after use.'},
    )

    banned_words_file: str = field(
        default=MISSING,
        metadata={
            'help': """
                Path to a text file with a list of banned words to block in the UI.
                Each row in the file is one word/phrase.
                User will receieve an alert and are asked to rephrase, if there is an exact match.
            """
        },
    )

    max_times_persona_use: int = field(
        default=0,
        metadata={
            'help': """
                Maximum number of times to allow a particular persona to be used.
                Default (0) mean no limit.
            """
        },
    )

    locations_file: str = field(
        default=MISSING,
        metadata={
            'help': """
                Path to a text file that keeps a list of locations that will be added
                to some of the curated personas (marked for needing persona).
            """
        },
    )

    search_server: str = field(
        default=MISSING, metadata={'help': 'Address to the search API.'}
    )

    num_passages_retrieved: int = field(
        default=5,
        metadata={'help': 'The number of documents to request from search API.'},
    )

    search_warning_turn: int = field(
        default=2,
        metadata={
            'help': 'The round that wizard may receive warning for using more search.'
        },
    )

    search_warning_threshold: int = field(
        default=2,
        metadata={
            'help': """
            The minimum number of times that wizard needs to use the search bar,
            at the rounds that we check for sending them a warning
            (warning is not send if wizard has used search more than this many times).
            """
        },
    )

    select_warning_turn: int = field(
        default=3,
        metadata={
            'help': 'The round that Wizard may receive warning to select more search results'
        },
    )

    select_warning_threshold: int = field(
        default=2,
        metadata={
            'help': """
            The minimum number of knowledge selections that wizard needs to have,
            at the round that we check for sending them a warning
            (warning is not send if Wizard has selected saerch results
            at least this many times so far).
            """
        },
    )


@register_mephisto_abstraction()
class WizardOfInternetBlueprint(ParlAIChatBlueprint):
    BLUEPRINT_TYPE = WIZARD_INTERNET_PARLAICHAT_BLUEPRINT
    ArgsClass = WizardOfInternetBlueprintArgs

    @classmethod
    def assert_task_args(
        cls, args: 'DictConfig', shared_state: 'SharedTaskState'
    ) -> None:
        """
        Ensure that arguments are properly configured to launch this task.
        """
        ParlAIChatBlueprint.assert_task_args(args=args, shared_state=shared_state)
        blueprint = args.get('blueprint')
        # Check search module is valid
        assert hasattr(blueprint, 'search_server'), 'Provide search API address.'

        assert hasattr(blueprint, 'use_personas_with_replacement')
        assert hasattr(shared_state, 'world_opt')
        assert 'personas' in shared_state.world_opt

        # Number of personas is enough for running without replacement
        if not blueprint.get('use_personas_with_replacement'):
            n_personas = len(shared_state.world_opt['personas'])
            n_conversations = blueprint.get('num_conversations')
            assert (
                n_personas >= n_conversations
            ), f'{n_personas} personas are not enought to use uniquely for {n_conversations} conversations.'

        # Make sure that we first show the warning for using search more often
        # to the wizard, and then the warning for selecting more sentences.
        assert blueprint.get('search_warning_turn') <= blueprint.get(
            'select_warning_turn'
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Feedforward neural network, as used in transformer implementation.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from parlai.core.opt import Opt


class TransformerFFN(nn.Module):
    """
    Implements the FFN part of the transformer.
    """

    def __init__(
        self,
        opt: Opt,
        dim: int = None,
        dim_hidden: int = None,
        relu_dropout: float = 0,
        activation: str = 'relu',
        **kwargs,
    ):
        super(TransformerFFN, self).__init__(**kwargs)

        def _default(val, default):
            """
            shorthand for explicit None check for optional arguments.
            """
            return val if val is not None else default

        dim = _default(dim, opt['embedding_size'])
        dim_hidden = _default(dim_hidden, opt['ffn_size'])

        self.opt = opt
        self.relu_dropout = nn.Dropout(p=relu_dropout)
        if activation == 'relu':
            self.nonlinear = F.relu
        elif activation == 'gelu':
            self.nonlinear = F.gelu
        else:
            raise ValueError(
                "Don't know how to handle --activation {}".format(activation)
            )
        self.lin1 = nn.Linear(dim, dim_hidden)
        self.lin2 = nn.Linear(dim_hidden, dim)
        nn.init.xavier_uniform_(self.lin1.weight)
        nn.init.xavier_uniform_(self.lin2.weight)
        # TODO: initialize biases to 0

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Forward pass.
        """
        x = self.nonlinear(self.lin1(x))
        x = self.relu_dropout(x)  # --relu-dropout
        x = self.lin2(x)
        return x



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = "squad"


class TestIndexTeacher(AutoTeacherTest):
    task = "squad:index"


class TestOpensquadTeacher(AutoTeacherTest):
    task = "squad:opensquad"


class TestFulldocTeacher(AutoTeacherTest):
    task = "squad:fulldoc"


class TestSentenceTeacher(AutoTeacherTest):
    task = "squad:sentence"


class TestFulldocsentenceTeacher(AutoTeacherTest):
    task = "squad:fulldocsentence"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from typing import List

import torch.jit

from parlai.utils.io import PathManager


def test_exported_model(scripted_model_file: str, inputs: List[str]):

    with PathManager.open(scripted_model_file, "rb") as f:
        scripted_module = torch.jit.load(f)

    print('\nGenerating given the scripted module:')
    context = []
    for input_ in inputs:
        print(' TEXT: ' + input_)
        context.append(input_)
        label = scripted_module('\n'.join(context))
        print("LABEL: " + label)
        context.append(label)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-smf',
        '--scripted-model-file',
        type=str,
        help='Where to load the scripted model checkpoint from',
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="hello world",
        help="Test input string to pass into the encoder of the scripted model. Separate lines with a pipe",
    )
    args = parser.parse_args()
    test_exported_model(
        scripted_model_file=args.scripted_model_file, inputs=args.input.split('|')
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import MultiTaskTeacher
from parlai.tasks.huggingface.agents import AbstractHuggingFaceTeacher
from copy import deepcopy


class AxTeacher(AbstractHuggingFaceTeacher):
    """
    Note: this is an evaluation dataset so it only has a test split
    Use a model trained on MulitNLI to produce predictions for this dataset.
    """

    hf_path = 'glue'
    hf_name = 'ax'
    hf_text_fields = ['premise', 'hypothesis']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': None, 'valid': None, 'test': 'test'}


class ColaTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'cola'
    hf_text_fields = ['sentence']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class MnliTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'mnli'
    hf_text_fields = ['premise', 'hypothesis']
    hf_label_field = 'label'
    hf_splits_mapping = {
        'train': 'train',
        'valid': 'validation_matched',
        'test': 'test_matched',
    }


class MnliMatchedTeacher(AbstractHuggingFaceTeacher):
    """
    Note: this is an evaluation dataset so it only has valid and test splits
    """

    hf_path = 'glue'
    hf_name = 'mnli_matched'
    hf_text_fields = ['premise', 'hypothesis']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': None, 'valid': 'validation', 'test': 'test'}


class MnliMismatchedTeacher(AbstractHuggingFaceTeacher):
    """
    Note: this is an evaluation dataset so it only has valid and test splits
    """

    hf_path = 'glue'
    hf_name = 'mnli_mismatched'
    hf_text_fields = ['premise', 'hypothesis']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': None, 'valid': 'validation', 'test': 'test'}


class MrpcTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'mrpc'
    hf_text_fields = ['sentence1', 'sentence2']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class QnliTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'qnli'
    hf_text_fields = ['sentence', 'question']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class QqpTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'qqp'
    hf_text_fields = ['question1', 'question2']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class RteTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'rte'
    hf_text_fields = ['sentence1', 'sentence2']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class Sst2Teacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'sst2'
    hf_text_fields = ['sentence']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class StsbTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'stsb'
    hf_text_fields = ['sentence1', 'sentence2']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class WnliTeacher(AbstractHuggingFaceTeacher):
    hf_path = 'glue'
    hf_name = 'wnli'
    hf_text_fields = ['sentence1', 'sentence2']
    hf_label_field = 'label'
    hf_splits_mapping = {'train': 'train', 'valid': 'validation', 'test': 'test'}


class GlueTeacher(MultiTaskTeacher):
    def __init__(self, opt, shared=None):
        glue_tasks = [
            'cola',
            'mnli',
            'mrpc',
            'qnli',
            'qqp',
            'rte',
            'sst2',
            'stsb',
            'wnli',
        ]
        glue_tasks = ['glue:' + t for t in glue_tasks]
        opt = deepcopy(opt)
        opt['task'] = ', '.join(glue_tasks)
        super().__init__(opt, shared)


class DefaultTeacher(GlueTeacher):
    pass
